from __future__ import annotations

import abc
import copy
import json
import logging
import re
import time
import uuid
from abc import ABC
from datetime import datetime, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Text,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import jsonpickle
import structlog
from dateutil import parser

import rasa.shared.utils.common
import rasa.shared.utils.io
from rasa.shared.constants import DOCS_URL_TRAINING_DATA
from rasa.shared.core.constants import (
    ACTION_LISTEN_NAME,
    ACTION_METADATA_EXECUTION_ERROR_MESSAGE,
    ACTION_METADATA_EXECUTION_SUCCESS,
    ACTION_NAME_SENDER_ID_CONNECTOR_STR,
    ACTION_SESSION_START_NAME,
    ENTITY_LABEL_SEPARATOR,
    ERROR_CODE_KEY,
    EXTERNAL_MESSAGE_PREFIX,
    IS_EXTERNAL,
    LOOP_INTERRUPTED,
    LOOP_NAME,
    USE_TEXT_FOR_FEATURIZATION,
)
from rasa.shared.exceptions import UnsupportedFeatureException
from rasa.shared.nlu.constants import (
    ACTION_NAME,
    ACTION_TEXT,
    COMMANDS,
    ENTITIES,
    ENTITY_ATTRIBUTE_CONFIDENCE,
    ENTITY_ATTRIBUTE_END,
    ENTITY_ATTRIBUTE_GROUP,
    ENTITY_ATTRIBUTE_ROLE,
    ENTITY_ATTRIBUTE_START,
    ENTITY_ATTRIBUTE_TEXT,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_VALUE,
    FULL_RETRIEVAL_INTENT_NAME_KEY,
    INTENT,
    INTENT_NAME_KEY,
    INTENT_RANKING_KEY,
    PREDICTED_CONFIDENCE_KEY,
    PROMPTS,
    TEXT,
)

if TYPE_CHECKING:
    from typing_extensions import TypedDict

    from rasa.shared.core.trackers import DialogueStateTracker

    EntityPrediction = TypedDict(
        "EntityPrediction",
        {
            ENTITY_ATTRIBUTE_TEXT: Text,  # type: ignore[misc]
            ENTITY_ATTRIBUTE_START: Optional[float],
            ENTITY_ATTRIBUTE_END: Optional[float],
            ENTITY_ATTRIBUTE_VALUE: Text,
            ENTITY_ATTRIBUTE_CONFIDENCE: float,
            ENTITY_ATTRIBUTE_TYPE: Text,
            ENTITY_ATTRIBUTE_GROUP: Optional[Text],
            ENTITY_ATTRIBUTE_ROLE: Optional[Text],
            "additional_info": Any,
        },
        total=False,
    )

    IntentPrediction = TypedDict(
        "IntentPrediction",
        {INTENT_NAME_KEY: Text, PREDICTED_CONFIDENCE_KEY: float},  # type: ignore[misc]
    )
    NLUPredictionData = TypedDict(
        "NLUPredictionData",
        {
            TEXT: Text,  # type: ignore[misc]
            INTENT: IntentPrediction,
            INTENT_RANKING_KEY: List[IntentPrediction],
            ENTITIES: List[EntityPrediction],
            "message_id": Optional[Text],
            "metadata": Dict,
            PROMPTS: Dict,
        },
        total=False,
    )
logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()

INVALID_DATETIME_ERROR_MESSAGE = "`anonymized_at` must be a datetime object."


def deserialise_events(serialized_events: List[Dict[Text, Any]]) -> List["Event"]:
    """Convert a list of dictionaries to a list of corresponding events.

    Example format:
        [{"event": "slot", "value": 5, "name": "my_slot"}]
    """
    deserialised = []

    for e in serialized_events:
        if "event" in e:
            event = Event.from_parameters(e)
            if event:
                deserialised.append(event)
            else:
                structlogger.warning("event.deserialization.failed")

    return deserialised


def deserialise_entities(entities: Union[Text, List[Any]]) -> List[Dict[Text, Any]]:
    if isinstance(entities, str):
        entities = json.loads(entities)

    return [e for e in entities if isinstance(e, dict)]


def format_message(
    text: Text, intent: Optional[Text], entities: Union[Text, List[Any]]
) -> Text:
    """Uses NLU parser information to generate a message with inline entity annotations.

    Arguments:
        text: text of the message
        intent: intent of the message
        entities: entities of the message

    Return:
        Message with entities annotated inline, e.g.
        `I am from [Berlin]{`"`entity`"`: `"`city`"`}`.
    """
    from rasa.shared.nlu.training_data import entities_parser
    from rasa.shared.nlu.training_data.formats.readerwriter import TrainingDataWriter

    message_from_md = entities_parser.parse_training_example(text, intent)
    deserialised_entities = deserialise_entities(entities)
    return TrainingDataWriter.generate_message(
        {"text": message_from_md.get(TEXT), "entities": deserialised_entities}
    )


def split_events(
    events: Iterable["Event"],
    event_type_to_split_on: Type["Event"],
    additional_splitting_conditions: Optional[Dict[Text, Any]] = None,
    include_splitting_event: bool = True,
) -> List[List["Event"]]:
    """Splits events according to an event type and condition.

    Examples:
        Splitting events according to the event type `ActionExecuted` and the
        `action_name` 'action_session_start' would look as follows:

        >> _events = split_events(
                        events,
                        ActionExecuted,
                        {"action_name": "action_session_start"},
                        True
                     )

    Args:
        events: Events to split.
        event_type_to_split_on: The event type to split on.
        additional_splitting_conditions: Additional event attributes to split on.
        include_splitting_event: Whether the events of the type on which the split
            is based should be included in the returned events.

    Returns:
        The split events.
    """
    sub_events = []
    current: List["Event"] = []

    def event_fulfills_splitting_condition(evt: "Event") -> bool:
        # event does not have the correct type
        if not isinstance(evt, event_type_to_split_on):
            return False

        # the type is correct and there are no further conditions
        if not additional_splitting_conditions:
            return True

        # there are further conditions - check those
        return all(
            getattr(evt, k, None) == v
            for k, v in additional_splitting_conditions.items()
        )

    for event in events:
        if event_fulfills_splitting_condition(event):
            if current:
                sub_events.append(current)

            current = []
            if include_splitting_event:
                current.append(event)
        else:
            current.append(event)

    if current:
        sub_events.append(current)

    return sub_events


def do_events_begin_with_session_start(events: List["Event"]) -> bool:
    """Determines whether `events` begins with a session start sequence.

    A session start sequence is a sequence of two events: an executed
    `action_session_start` as well as a logged `session_started`.

    Args:
        events: The events to inspect.

    Returns:
        Whether `events` begins with a session start sequence.
    """
    if len(events) < 2:
        return False

    first = events[0]
    second = events[1]

    # We are not interested in specific metadata or timestamps. Action name and event
    # type are sufficient for this check
    return (
        isinstance(first, ActionExecuted)
        and first.action_name == ACTION_SESSION_START_NAME
        and isinstance(second, SessionStarted)
    )


def remove_parse_data(event: Dict[Text, Any]) -> Dict[Text, Any]:
    """Reduce event details to the minimum necessary to be structlogged.

    Deletes the parse_data key from the event if it exists.

    Args:
        event: The event to be reduced.

    Returns:
        A reduced copy of the event.
    """
    reduced_event = copy.deepcopy(event)
    if "parse_data" in reduced_event:
        del reduced_event["parse_data"]
    return reduced_event


E = TypeVar("E", bound="Event")


class Event(ABC):
    """Describes events in conversation and how the affect the conversation state.

    Immutable representation of everything which happened during a conversation of the
    user with the assistant. Tells the `rasa.shared.core.trackers.DialogueStateTracker`
    how to update its state as the events occur.
    """

    type_name = "event"

    def __init__(
        self,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        self.timestamp = timestamp or time.time()
        self.metadata = metadata or {}

    def __ne__(self, other: Any) -> bool:
        # Not strictly necessary, but to avoid having both x==y and x!=y
        # True at the same time
        return not (self == other)

    @abc.abstractmethod
    def as_story_string(self) -> Optional[Text]:
        """Returns the event as story string.

        Returns:
            textual representation of the event or None.
        """
        # Every class should implement this
        raise NotImplementedError

    @staticmethod
    def from_story_string(
        event_name: Text,
        parameters: Dict[Text, Any],
        default: Optional[Type["Event"]] = None,
    ) -> Optional[List["Event"]]:
        event_class = Event.resolve_by_type(event_name, default)

        if not event_class:
            return None

        return event_class._from_story_string(parameters)

    @staticmethod
    def from_parameters(
        parameters: Dict[Text, Any], default: Optional[Type["Event"]] = None
    ) -> Optional["Event"]:
        event_name = parameters.get("event")
        if event_name is None:
            return None

        event_class: Optional[Type[Event]] = Event.resolve_by_type(event_name, default)
        if not event_class:
            return None

        return event_class._from_parameters(parameters)

    @classmethod
    def _from_story_string(
        cls: Type[E], parameters: Dict[Text, Any]
    ) -> Optional[List[E]]:
        """Called to convert a parsed story line into an event."""
        return [cls(parameters.get("timestamp"), parameters.get("metadata"))]

    def as_dict(self) -> Dict[Text, Any]:
        d = {"event": self.type_name, "timestamp": self.timestamp}

        if self.metadata:
            d["metadata"] = self.metadata

        return d

    def fingerprint(self) -> Text:
        """Returns a unique hash for the event which is stable across python runs.

        Returns:
            fingerprint of the event
        """
        data = self.as_dict()
        del data["timestamp"]
        return rasa.shared.utils.io.get_dictionary_fingerprint(data)

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> Optional["Event"]:
        """Called to convert a dictionary of parameters to a single event.

        By default uses the same implementation as the story line
        conversation ``_from_story_string``. But the subclass might
        decide to handle parameters differently if the parsed parameters
        don't origin from a story file.
        """
        result = cls._from_story_string(parameters)
        if len(result) > 1:
            logger.warning(
                f"Event from parameters called with parameters "
                f"for multiple events. This is not supported, "
                f"only the first event will be returned. "
                f"Parameters: {parameters}"
            )
        return result[0] if result else None

    @staticmethod
    def resolve_by_type(
        type_name: Text, default: Optional[Type["Event"]] = None
    ) -> Optional[Type["Event"]]:
        """Returns a slots class by its type name."""
        for cls in rasa.shared.utils.common.all_subclasses(Event):
            if cls.type_name == type_name:
                return cls
        if type_name == "topic":
            return None  # backwards compatibility to support old TopicSet evts
        elif default is not None:
            return default
        else:
            raise ValueError(f"Unknown event name '{type_name}'.")

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state.

        Args:
            tracker: The current conversation state.
        """
        pass

    @abc.abstractmethod
    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        # Every class should implement this
        raise NotImplementedError()

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return f"{self.__class__.__name__}()"


class AlwaysEqualEventMixin(Event, ABC):
    """Class to deduplicate common behavior for events without additional attributes."""

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, self.__class__):
            return NotImplemented

        return True


class SkipEventInMDStoryMixin(Event, ABC):
    """Skips the visualization of an event in Markdown stories."""

    def as_story_string(self) -> None:
        """Returns the event as story string.

        Returns:
            None, as this event should not appear inside the story.
        """
        return


class UserUttered(Event):
    """The user has said something to the bot.

    As a side effect a new `Turn` will be created in the `Tracker`.
    """

    type_name = "user"

    def __init__(
        self,
        text: Optional[Text] = None,
        intent: Optional[Dict] = None,
        entities: Optional[List[Dict]] = None,
        parse_data: Optional["NLUPredictionData"] = None,
        timestamp: Optional[float] = None,
        input_channel: Optional[Text] = None,
        message_id: Optional[Text] = None,
        metadata: Optional[Dict] = None,
        use_text_for_featurization: Optional[bool] = None,
        anonymized_at: Optional[float] = None,
    ) -> None:
        """Creates event for incoming user message.

        Args:
            text: Text of user message.
            intent: Intent prediction of user message.
            entities: Extracted entities.
            parse_data: Detailed NLU parsing result for message.
            timestamp: When the event was created.
            metadata: Additional event metadata.
            input_channel: Which channel the user used to send message.
            message_id: Unique ID for message.
            use_text_for_featurization: `True` if the message's text was used to predict
                next action. `False` if the message's intent was used.
            anonymized_at: When the event was anonymized in the tracker store.

        """
        self.text = text
        self.intent = intent if intent else {}
        self.entities = entities if entities else []
        self.input_channel = input_channel
        self.message_id = message_id

        super().__init__(timestamp, metadata)

        # The featurization is set by the policies during prediction time using a
        # `DefinePrevUserUtteredFeaturization` event.
        self.use_text_for_featurization = use_text_for_featurization
        # define how this user utterance should be featurized
        if self.text and not self.intent_name:
            # happens during training
            self.use_text_for_featurization = True
        elif self.intent_name and not self.text:
            # happens during training
            self.use_text_for_featurization = False

        self.parse_data: "NLUPredictionData" = {
            INTENT: self.intent,  # type: ignore[misc]
            # Copy entities so that changes to `self.entities` don't affect
            # `self.parse_data` and hence don't get persisted
            ENTITIES: self.entities.copy(),
            TEXT: self.text,
            "message_id": self.message_id,
            "metadata": self.metadata,
        }
        if parse_data:
            self.parse_data.update(**parse_data)

        self._anonymized_at: Optional[datetime] = (
            datetime.fromtimestamp(anonymized_at, tz=timezone.utc)
            if anonymized_at is not None
            else None
        )

    @property
    def anonymized_at(self) -> Optional[datetime]:
        """Returns the time when the event was anonymized in the tracker store.

        If the event was not anonymized, it returns None.
        """
        return self._anonymized_at

    @anonymized_at.setter
    def anonymized_at(self, value: Optional[datetime]) -> None:
        """Sets the time when the event was anonymized in the tracker store."""
        if value is not None and not isinstance(value, datetime):
            raise ValueError(INVALID_DATETIME_ERROR_MESSAGE)
        self._anonymized_at = value

    @staticmethod
    def _from_parse_data(
        text: Text,
        parse_data: "NLUPredictionData",
        timestamp: Optional[float] = None,
        input_channel: Optional[Text] = None,
        message_id: Optional[Text] = None,
        metadata: Optional[Dict] = None,
        anonymized_at: Optional[float] = None,
    ) -> "UserUttered":
        return UserUttered(
            text,
            parse_data.get(INTENT),
            parse_data.get(ENTITIES, []),
            parse_data,
            timestamp,
            input_channel,
            message_id,
            metadata,
            anonymized_at,
        )

    def __hash__(self) -> int:
        """Returns unique hash of object."""
        return hash(json.dumps(self.as_sub_state()))

    @property
    def has_triggered_error(self) -> bool:
        return len(self.error_commands) > 0

    @property
    def error_commands(self) -> List[Dict[Text, Any]]:
        from rasa.dialogue_understanding.commands import ErrorCommand

        return [c for c in self.commands if c["command"] == ErrorCommand.command()]

    @property
    def commands(self) -> List[Dict[str, Any]]:
        """Returns commands included in the message."""
        if COMMANDS in self.parse_data and isinstance(
            self.parse_data[COMMANDS],  # type: ignore[literal-required]
            list,
        ):
            return self.parse_data[COMMANDS]  # type: ignore[literal-required]
        return []

    @property
    def intent_name(self) -> Optional[Text]:
        """Returns intent name or `None` if no intent."""
        return self.intent.get(INTENT_NAME_KEY)

    @property
    def full_retrieval_intent_name(self) -> Optional[Text]:
        """Returns full retrieval intent name or `None` if no retrieval intent."""
        return self.intent.get(FULL_RETRIEVAL_INTENT_NAME_KEY)

    # Note that this means two UserUttered events with the same text, intent
    # and entities but _different_ timestamps will be considered equal.
    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, UserUttered):
            return NotImplemented

        return (
            self.text,
            self.intent_name,
            [
                jsonpickle.encode(sorted(ent)) for ent in self.entities
            ],  # TODO: test? Or fix in regex_message_handler?
        ) == (
            other.text,
            other.intent_name,
            [jsonpickle.encode(sorted(ent)) for ent in other.entities],
        )

    def __str__(self) -> Text:
        """Returns text representation of event."""
        entities = ""
        if self.entities:
            entities_list = [
                f"{entity[ENTITY_ATTRIBUTE_VALUE]} "
                f"(Type: {entity[ENTITY_ATTRIBUTE_TYPE]}, "
                f"Role: {entity.get(ENTITY_ATTRIBUTE_ROLE)}, "
                f"Group: {entity.get(ENTITY_ATTRIBUTE_GROUP)})"
                for entity in self.entities
            ]
            entities = f", entities: {', '.join(entities_list)}"

        return (
            f"UserUttered(text: {self.text}, intent: {self.intent_name}"
            f"{entities}"
            f", use_text_for_featurization: {self.use_text_for_featurization})"
        )

    def __repr__(self) -> Text:
        """Returns text representation of event for debugging."""
        return (
            f"UserUttered('{self.text}', "
            f"'{self.intent_name}', "
            f"{json.dumps(self.entities)})"
        )

    @staticmethod
    def empty() -> "UserUttered":
        return UserUttered(None)

    def is_empty(self) -> bool:
        return not self.text and not self.intent_name and not self.entities

    def as_dict(self) -> Dict[Text, Any]:
        _dict = super().as_dict()
        _dict.update(
            {
                "text": self.text,
                "parse_data": self.parse_data,
                "input_channel": getattr(self, "input_channel", None),
                "message_id": getattr(self, "message_id", None),
                "metadata": self.metadata,
                "anonymized_at": self.anonymized_at.timestamp()
                if self.anonymized_at
                else None,
            }
        )
        return _dict

    def as_sub_state(self) -> Dict[Text, Union[None, Text, List[Optional[Text]]]]:
        """Turns a UserUttered event into features.

        The substate contains information about entities, intent and text of the
        `UserUttered` event.

        Returns:
            a dictionary with intent name, text and entities
        """
        entities = [entity.get(ENTITY_ATTRIBUTE_TYPE) for entity in self.entities]
        entities.extend(
            (
                f"{entity.get(ENTITY_ATTRIBUTE_TYPE)}{ENTITY_LABEL_SEPARATOR}"
                f"{entity.get(ENTITY_ATTRIBUTE_ROLE)}"
            )
            for entity in self.entities
            if ENTITY_ATTRIBUTE_ROLE in entity
        )
        entities.extend(
            (
                f"{entity.get(ENTITY_ATTRIBUTE_TYPE)}{ENTITY_LABEL_SEPARATOR}"
                f"{entity.get(ENTITY_ATTRIBUTE_GROUP)}"
            )
            for entity in self.entities
            if ENTITY_ATTRIBUTE_GROUP in entity
        )

        out: Dict[Text, Union[None, Text, List[Optional[Text]]]] = {}
        # During training we expect either intent_name or text to be set.
        # During prediction both will be set.
        if self.text and (
            self.use_text_for_featurization or self.use_text_for_featurization is None
        ):
            out[TEXT] = self.text
        if self.intent_name and not self.use_text_for_featurization:
            out[INTENT] = self.intent_name
        # don't add entities for e2e utterances
        if entities and not self.use_text_for_featurization:
            out[ENTITIES] = entities

        return out

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["UserUttered"]]:
        try:
            return [
                cls._from_parse_data(
                    parameters.get("text"),
                    parameters.get("parse_data"),
                    parameters.get("timestamp"),
                    parameters.get("input_channel"),
                    parameters.get("message_id"),
                    parameters.get("metadata"),
                    parameters.get("anonymized_at"),
                )
            ]
        except KeyError as e:
            raise ValueError(f"Failed to parse bot uttered event. {e}")

    def _entity_string(self) -> Text:
        if self.entities:
            return json.dumps(
                {
                    entity[ENTITY_ATTRIBUTE_TYPE]: entity[ENTITY_ATTRIBUTE_VALUE]
                    for entity in self.entities
                },
                ensure_ascii=False,
            )
        return ""

    def as_story_string(self, e2e: bool = False) -> Text:
        """Return event as string for Markdown training format.

        Args:
            e2e: `True` if the the event should be printed in the format for
                end-to-end conversation tests.

        Returns:
            Event as string.
        """
        if self.use_text_for_featurization and not e2e:
            raise UnsupportedFeatureException(
                f"Printing end-to-end user utterances is not supported in the "
                f"Markdown training format. Please use the YAML training data format "
                f"instead. Please see {DOCS_URL_TRAINING_DATA} for more information."
            )

        if e2e:
            text_with_entities = format_message(
                self.text or "", self.intent_name, self.entities
            )

            intent_prefix = f"{self.intent_name}: " if self.intent_name else ""
            return f"{intent_prefix}{text_with_entities}"

        return f"{self.intent_name or ''}{self._entity_string()}"

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to tracker. See docstring of `Event`."""
        tracker.latest_message = self
        tracker.clear_followup_action()

    @staticmethod
    def create_external(
        intent_name: Text,
        entity_list: Optional[List[Dict[Text, Any]]] = None,
        input_channel: Optional[Text] = None,
    ) -> "UserUttered":
        return UserUttered(
            text=f"{EXTERNAL_MESSAGE_PREFIX}{intent_name}",
            intent={INTENT_NAME_KEY: intent_name},
            metadata={IS_EXTERNAL: True},
            entities=entity_list or [],
            input_channel=input_channel,
        )


class DefinePrevUserUtteredFeaturization(SkipEventInMDStoryMixin):
    """Stores information whether action was predicted based on text or intent."""

    type_name = "user_featurization"

    def __init__(
        self,
        use_text_for_featurization: bool,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event.

        Args:
            use_text_for_featurization: `True` if message text was used to predict
                action. `False` if intent was used.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        super().__init__(timestamp, metadata)
        self.use_text_for_featurization = use_text_for_featurization

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return f"DefinePrevUserUtteredFeaturization({self.use_text_for_featurization})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.use_text_for_featurization)

    @classmethod
    def _from_parameters(
        cls, parameters: Dict[Text, Any]
    ) -> "DefinePrevUserUtteredFeaturization":
        return DefinePrevUserUtteredFeaturization(
            parameters.get(USE_TEXT_FOR_FEATURIZATION),
            parameters.get("timestamp"),
            parameters.get("metadata"),
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update({USE_TEXT_FOR_FEATURIZATION: self.use_text_for_featurization})
        return d

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state.

        Args:
            tracker: The current conversation state.
        """
        if tracker.latest_action_name != ACTION_LISTEN_NAME:
            # featurization belong only to the last user message
            # a user message is always followed by action listen
            return

        if not tracker.latest_message:
            return

        # update previous user message's featurization based on this event
        tracker.latest_message.use_text_for_featurization = (
            self.use_text_for_featurization
        )

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, DefinePrevUserUtteredFeaturization):
            return NotImplemented

        return self.use_text_for_featurization == other.use_text_for_featurization


class EntitiesAdded(SkipEventInMDStoryMixin):
    """Event that is used to add extracted entities to the tracker state."""

    type_name = "entities"

    def __init__(
        self,
        entities: List[Dict[Text, Any]],
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Initializes event.

        Args:
            entities: Entities extracted from previous user message. This can either
                be done by NLU components or end-to-end policy predictions.
            timestamp: the timestamp
            metadata: some optional metadata
        """
        super().__init__(timestamp, metadata)
        self.entities = entities

    def __str__(self) -> Text:
        """Returns the string representation of the event."""
        entity_str = [e[ENTITY_ATTRIBUTE_TYPE] for e in self.entities]
        return f"{self.__class__.__name__}({entity_str})"

    def __hash__(self) -> int:
        """Returns the hash value of the event."""
        return hash(json.dumps(self.entities))

    def __eq__(self, other: Any) -> bool:
        """Compares this event with another event."""
        if not isinstance(other, EntitiesAdded):
            return NotImplemented

        return self.entities == other.entities

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "EntitiesAdded":
        return EntitiesAdded(
            parameters.get(ENTITIES),
            parameters.get("timestamp"),
            parameters.get("metadata"),
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Converts the event into a dict.

        Returns:
            A dict that represents this event.
        """
        d = super().as_dict()
        d.update({ENTITIES: self.entities})
        return d

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state.

        Args:
            tracker: The current conversation state.
        """
        if tracker.latest_action_name != ACTION_LISTEN_NAME:
            # entities belong only to the last user message
            # a user message always comes after action listen
            return

        if not tracker.latest_message:
            return

        for entity in self.entities:
            if entity not in tracker.latest_message.entities:
                tracker.latest_message.entities.append(entity)


class BotUttered(SkipEventInMDStoryMixin):
    """The bot has said something to the user.

    This class is not used in the story training as it is contained in the

    ``ActionExecuted`` class. An entry is made in the ``Tracker``.
    """

    type_name = "bot"

    def __init__(
        self,
        text: Optional[Text] = None,
        data: Optional[Dict] = None,
        metadata: Optional[Dict[Text, Any]] = None,
        timestamp: Optional[float] = None,
        anonymized_at: Optional[float] = None,
    ) -> None:
        """Creates event for a bot response.

        Args:
            text: Plain text which bot responded with.
            data: Additional data for more complex utterances (e.g. buttons).
            timestamp: When the event was created.
            metadata: Additional event metadata.
            anonymized_at: When the event was anonymized in the tracker store.
        """
        self.text = text
        self.data = data or {}
        self._anonymized_at: Optional[datetime] = (
            datetime.fromtimestamp(anonymized_at, tz=timezone.utc)
            if anonymized_at is not None
            else None
        )
        super().__init__(timestamp, metadata)

    @property
    def anonymized_at(self) -> Optional[datetime]:
        """Returns the time when the event was anonymized in the tracker store.

        If the event was not anonymized, it returns None.
        """
        return self._anonymized_at

    @anonymized_at.setter
    def anonymized_at(self, value: Optional[datetime]) -> None:
        """Sets the time when the event was anonymized in the tracker store."""
        if value is not None and not isinstance(value, datetime):
            raise ValueError(INVALID_DATETIME_ERROR_MESSAGE)
        self._anonymized_at = value

    def __members(self) -> Tuple[Optional[Text], Text, Text]:
        data_no_nones = {k: v for k, v in self.data.items() if v is not None}
        meta_no_nones = {k: v for k, v in self.metadata.items() if v is not None}
        return (
            self.text,
            jsonpickle.encode(data_no_nones),
            jsonpickle.encode(meta_no_nones),
        )

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.__members())

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, BotUttered):
            return NotImplemented

        return self.__members() == other.__members()

    def _clean_up_metadata(self) -> Dict[str, Any]:
        """Removes search_results metadata key from the metadata.

        This is intended to prevent increasing the string representation
        character length of the bot event.
        """
        from rasa.core.policies.enterprise_search_policy import (
            SEARCH_RESULTS_METADATA_KEY,
        )

        metadata = copy.deepcopy(self.metadata)

        if SEARCH_RESULTS_METADATA_KEY in self.metadata:
            metadata.pop(SEARCH_RESULTS_METADATA_KEY)
            structlogger.debug(
                "search_results.metadata.removed",
                event_info="Removed search_results metadata key only "
                "from the string representation of the bot event.",
            )

        return metadata

    def __str__(self) -> Text:
        """Returns text representation of event."""
        metadata = self._clean_up_metadata()

        return "BotUttered(text: {}, data: {}, metadata: {})".format(
            self.text, json.dumps(self.data), json.dumps(metadata)
        )

    def __repr__(self) -> Text:
        """Returns text representation of event for debugging."""
        metadata = self._clean_up_metadata()

        return "BotUttered('{}', {}, {}, {})".format(
            self.text, json.dumps(self.data), json.dumps(metadata), self.timestamp
        )

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker.latest_bot_utterance = self

    def message(self) -> Dict[Text, Any]:
        """Return the complete message as a dictionary."""
        m = self.data.copy()
        m["text"] = self.text
        m["timestamp"] = self.timestamp
        m.update(self.metadata)

        if m.get("image") == m.get("attachment"):
            # we need this as there is an oddity we introduced a while ago where
            # we automatically set the attachment to the image. to not break
            # any persisted events we kept that, but we need to make sure that
            # the message contains the image only once
            m["attachment"] = None

        return m

    @staticmethod
    def empty() -> "BotUttered":
        """Creates an empty bot utterance."""
        return BotUttered()

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update(
            {
                "text": self.text,
                "data": self.data,
                "metadata": self.metadata,
                "anonymized_at": self.anonymized_at.timestamp()
                if self.anonymized_at
                else None,
            }
        )
        return d

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "BotUttered":
        try:
            return BotUttered(
                parameters.get("text"),
                parameters.get("data"),
                parameters.get("metadata"),
                parameters.get("timestamp"),
                parameters.get("anonymized_at"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse bot uttered event. {e}")


class SlotSet(Event):
    """The user has specified their preference for the value of a `slot`.

    Every slot has a name and a value. This event can be used to set a
    value for a slot on a conversation.

    As a side effect the `Tracker`'s slots will be updated so
    that `tracker.slots[key]=value`.
    """

    type_name = "slot"

    def __init__(
        self,
        key: Text,
        value: Optional[Any] = None,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
        filled_by: Optional[str] = None,
        anonymized_at: Optional[float] = None,
    ) -> None:
        """Creates event to set slot.

        Args:
            key: Name of the slot which is set.
            value: Value to which slot is set.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.key = key
        self.value = value
        self._filled_by = filled_by
        self._anonymized_at: Optional[datetime] = (
            datetime.fromtimestamp(anonymized_at, tz=timezone.utc)
            if anonymized_at is not None
            else None
        )
        super().__init__(timestamp, metadata)

    @property
    def anonymized_at(self) -> Optional[datetime]:
        """Returns the time when the event was anonymized in the tracker store.

        If the event was not anonymized, it returns None.
        """
        return self._anonymized_at

    @anonymized_at.setter
    def anonymized_at(self, value: Optional[datetime]) -> None:
        """Sets the time when the event was anonymized in the tracker store."""
        if value is not None and not isinstance(value, datetime):
            raise ValueError(INVALID_DATETIME_ERROR_MESSAGE)
        self._anonymized_at = value

    def __repr__(self) -> Text:
        """Returns text representation of event."""
        return f"SlotSet(key: {self.key}, value: {self.value})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash((self.key, jsonpickle.encode(self.value)))

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, SlotSet):
            return NotImplemented

        return (self.key, self.value) == (other.key, other.value)

    @property
    def filled_by(self) -> Optional[str]:
        return self._filled_by

    @filled_by.setter
    def filled_by(self, value: str) -> None:
        self._filled_by = value

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        props = json.dumps({self.key: self.value}, ensure_ascii=False)
        return f"{self.type_name}{props}"

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["SlotSet"]]:
        slots = []
        for slot_key, slot_val in parameters.items():
            slots.append(SlotSet(slot_key, slot_val))

        if slots:
            return slots
        else:
            return None

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update(
            {
                "name": self.key,
                "value": self.value,
                "filled_by": self.filled_by,
                "anonymized_at": self.anonymized_at.timestamp()
                if self.anonymized_at
                else None,
            }
        )
        return d

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "SlotSet":
        try:
            return SlotSet(
                parameters.get("name"),
                parameters.get("value"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
                filled_by=parameters.get("filled_by"),
                anonymized_at=parameters.get("anonymized_at"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse set slot event. {e}")

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker._set_slot(self.key, self.value, self.filled_by)


class Restarted(AlwaysEqualEventMixin):
    """Conversation should start over & history wiped.

    Instead of deleting all events, this event can be used to reset the
    trackers state (e.g. ignoring any past user messages & resetting all
    the slots).
    """

    type_name = "restart"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124312)

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        return self.type_name

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Resets the tracker and triggers a followup `ActionSessionStart`."""
        tracker._reset()
        tracker.trigger_followup_action(ACTION_SESSION_START_NAME)


class UserUtteranceReverted(AlwaysEqualEventMixin):
    """Bot reverts everything until before the most recent user message.

    The bot will revert all events after the latest `UserUttered`, this
    also means that the last event on the tracker is usually `action_listen`
    and the bot is waiting for a new user message.
    """

    type_name = "rewind"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124315)

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        return self.type_name

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker._reset()
        tracker.replay_events()


class AllSlotsReset(AlwaysEqualEventMixin):
    """All Slots are reset to their initial values.

    If you want to keep the dialogue history and only want to reset the
    slots, you can use this event to set all the slots to their initial
    values.
    """

    type_name = "reset_slots"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124316)

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        return self.type_name

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker._reset_slots()


class DialogueStackUpdated(Event):
    """Update the stack of a conversation."""

    type_name = "stack"

    def __init__(
        self,
        update: str,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates an event which updates the stack on a tracker.

        Args:
            update: The update to the stack.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.update = update
        super().__init__(timestamp, metadata)

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.update)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, DialogueStackUpdated):
            return NotImplemented

        return self.update == other.update

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return f"DialogueStackUpdate(update: {self.update})"

    def __repr__(self) -> Text:
        """Returns text representation of event for debugging."""
        return f'DialogueStackUpdate("""{self.update}""")'

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        props = json.dumps({"update": self.update})
        return f"{self.type_name}{props}"

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["DialogueStackUpdated"]]:
        return [
            DialogueStackUpdated(
                parameters.get("update"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        ]

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update({"update": self.update})
        return d

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker.apply_stack_update(self.update)

    def update_as_json(self) -> List[Dict[Text, Any]]:
        """Return the update as a JsonPatch object."""
        return json.loads(self.update)


class ReminderScheduled(Event):
    """Schedules the asynchronous triggering of a user intent at a given time.

    The triggered intent can include entities if needed.
    """

    type_name = "reminder"

    def __init__(
        self,
        intent: Text,
        trigger_date_time: datetime,
        entities: Optional[List[Dict]] = None,
        name: Optional[Text] = None,
        kill_on_user_message: bool = True,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates the reminder.

        Args:
            intent: Name of the intent to be triggered.
            trigger_date_time: Date at which the execution of the action
                should be triggered (either utc or with tz).
            name: ID of the reminder. If there are multiple reminders with
                 the same id only the last will be run.
            entities: Entities that should be supplied together with the
                 triggered intent.
            kill_on_user_message: ``True`` means a user message before the
                 trigger date will abort the reminder.
            timestamp: Creation date of the event.
            metadata: Optional event metadata.
        """
        self.intent = intent
        self.entities = entities
        self.trigger_date_time = trigger_date_time
        self.kill_on_user_message = kill_on_user_message
        self.name = name if name is not None else str(uuid.uuid1())
        super().__init__(timestamp, metadata)

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(
            (
                self.intent,
                self.entities,
                self.trigger_date_time.isoformat(),
                self.kill_on_user_message,
                self.name,
            )
        )

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, ReminderScheduled):
            return NotImplemented

        return self.name == other.name

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return (
            f"ReminderScheduled(intent: {self.intent}, "
            f"trigger_date: {self.trigger_date_time}, "
            f"entities: {self.entities}, name: {self.name})"
        )

    def scheduled_job_name(self, sender_id: Text) -> Text:
        return (
            f"[{hash(self.name)},{hash(self.intent)},{hash(str(self.entities))}]"
            f"{ACTION_NAME_SENDER_ID_CONNECTOR_STR}"
            f"{sender_id}"
        )

    def _properties(self) -> Dict[Text, Any]:
        return {
            "intent": self.intent,
            "date_time": self.trigger_date_time.isoformat(),
            "entities": self.entities,
            "name": self.name,
            "kill_on_user_msg": self.kill_on_user_message,
        }

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        props = json.dumps(self._properties())
        return f"{self.type_name}{props}"

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update(self._properties())
        return d

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["ReminderScheduled"]]:
        trigger_date_time = parser.parse(parameters.get("date_time"))

        return [
            ReminderScheduled(
                parameters.get("intent"),
                trigger_date_time,
                parameters.get("entities"),
                name=parameters.get("name"),
                kill_on_user_message=parameters.get("kill_on_user_msg", True),
                timestamp=parameters.get("timestamp"),
                metadata=parameters.get("metadata"),
            )
        ]


class ReminderCancelled(Event):
    """Cancel certain jobs."""

    type_name = "cancel_reminder"

    def __init__(
        self,
        name: Optional[Text] = None,
        intent: Optional[Text] = None,
        entities: Optional[List[Dict]] = None,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates a ReminderCancelled event.

        If all arguments are `None`, this will cancel all reminders.
        are to be cancelled. If no arguments are supplied, this will cancel all
        reminders.

        Args:
            name: Name of the reminder to be cancelled.
            intent: Intent name that is to be used to identify the reminders to be
                cancelled.
            entities: Entities that are to be used to identify the reminders to be
                cancelled.
            timestamp: Optional timestamp.
            metadata: Optional event metadata.
        """
        self.name = name
        self.intent = intent
        self.entities = entities
        super().__init__(timestamp, metadata)

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash((self.name, self.intent, str(self.entities)))

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, ReminderCancelled):
            return NotImplemented

        return hash(self) == hash(other)

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return (
            f"ReminderCancelled(name: {self.name}, intent: {self.intent}, "
            f"entities: {self.entities})"
        )

    def cancels_job_with_name(self, job_name: Text, sender_id: Text) -> bool:
        """Determines if this event should cancel the job with the given name.

        Args:
            job_name: Name of the job to be tested.
            sender_id: The `sender_id` of the tracker.

        Returns:
            `True`, if this `ReminderCancelled` event should cancel the job with the
            given name, and `False` otherwise.
        """
        match = re.match(
            rf"^\[([\d\-]*),([\d\-]*),([\d\-]*)\]"
            rf"({re.escape(ACTION_NAME_SENDER_ID_CONNECTOR_STR)}"
            rf"{re.escape(sender_id)})",
            job_name,
        )
        if not match:
            return False
        name_hash, intent_hash, entities_hash = match.group(1, 2, 3)

        # Cancel everything unless names/intents/entities are given to
        # narrow it down.
        return (
            ((not self.name) or self._matches_name_hash(name_hash))
            and ((not self.intent) or self._matches_intent_hash(intent_hash))
            and ((not self.entities) or self._matches_entities_hash(entities_hash))
        )

    def _matches_name_hash(self, name_hash: Text) -> bool:
        return str(hash(self.name)) == name_hash

    def _matches_intent_hash(self, intent_hash: Text) -> bool:
        return str(hash(self.intent)) == intent_hash

    def _matches_entities_hash(self, entities_hash: Text) -> bool:
        return str(hash(str(self.entities))) == entities_hash

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        props = json.dumps(
            {"name": self.name, "intent": self.intent, "entities": self.entities}
        )
        return f"{self.type_name}{props}"

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["ReminderCancelled"]]:
        return [
            ReminderCancelled(
                parameters.get("name"),
                parameters.get("intent"),
                parameters.get("entities"),
                timestamp=parameters.get("timestamp"),
                metadata=parameters.get("metadata"),
            )
        ]


class ActionReverted(AlwaysEqualEventMixin):
    """Bot undoes its last action.

    The bot reverts everything until before the most recent action.
    This includes the action itself, as well as any events that
    action created, like set slot events - the bot will now
    predict a new action using the state before the most recent
    action.
    """

    type_name = "undo"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124318)

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        return self.type_name

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker._reset()
        tracker.replay_events()


class StoryExported(Event):
    """Story should get dumped to a file."""

    type_name = "export"

    def __init__(
        self,
        path: Optional[Text] = None,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event about story exporting.

        Args:
            path: Path to which story was exported to.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.path = path
        super().__init__(timestamp, metadata)

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124319)

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["StoryExported"]]:
        return [
            StoryExported(
                parameters.get("path"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        ]

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        return self.type_name

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        if self.path:
            tracker.export_stories_to_file(self.path)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, StoryExported):
            return NotImplemented

        return self.path == other.path


class FollowupAction(Event):
    """Enqueue a followup action."""

    type_name = "followup"

    def __init__(
        self,
        name: Text,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates an event which forces the model to run a certain action next.

        Args:
            name: Name of the action to run.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.action_name = name
        super().__init__(timestamp, metadata)

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.action_name)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, FollowupAction):
            return NotImplemented

        return self.action_name == other.action_name

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return f"FollowupAction(action: {self.action_name})"

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        props = json.dumps({"name": self.action_name})
        return f"{self.type_name}{props}"

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["FollowupAction"]]:
        return [
            FollowupAction(
                parameters.get("name"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        ]

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update({"name": self.action_name})
        return d

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker.trigger_followup_action(self.action_name)


class ConversationPaused(AlwaysEqualEventMixin):
    """Ignore messages from the user to let a human take over.

    As a side effect the `Tracker`'s `paused` attribute will
    be set to `True`.
    """

    type_name = "pause"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124313)

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        return str(self)

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker._paused = True


class ConversationResumed(AlwaysEqualEventMixin):
    """Bot takes over conversation.

    Inverse of `PauseConversation`. As a side effect the `Tracker`'s
    `paused` attribute will be set to `False`.
    """

    type_name = "resume"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124314)

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        return self.type_name

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker._paused = False


class ActionExecuted(Event):
    """An operation describes an action taken + its result.

    It comprises an action and a list of events. operations will be appended
    to the latest `Turn`` in `Tracker.turns`.
    """

    type_name = "action"

    def __init__(
        self,
        action_name: Optional[Text] = None,
        policy: Optional[Text] = None,
        confidence: Optional[float] = None,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict] = None,
        action_text: Optional[Text] = None,
        hide_rule_turn: bool = False,
    ) -> None:
        """Creates event for a successful event execution.

        Args:
            action_name: Name of the action which was executed. `None` if it was an
                end-to-end prediction.
            policy: Policy which predicted action.
            confidence: Confidence with which policy predicted action.
            timestamp: When the event was created.
            metadata: Additional event metadata.
            action_text: In case it's an end-to-end action prediction, the text which
                was predicted.
            hide_rule_turn: If `True`, this action should be hidden in the dialogue
                history created for ML-based policies.
        """
        self.action_name = action_name
        self.policy = policy
        self.confidence = confidence
        self.unpredictable = False
        self.action_text = action_text
        self.hide_rule_turn = hide_rule_turn

        if self.action_name is None and self.action_text is None:
            raise ValueError(
                "Both the name of the action and the end-to-end "
                "predicted text are missing. "
                "The `ActionExecuted` event cannot be initialised."
            )

        super().__init__(timestamp, metadata)

    def __members__(self) -> Tuple[Optional[Text], Optional[Text], Text]:
        items_to_ignore = [
            ACTION_METADATA_EXECUTION_SUCCESS,
            ACTION_METADATA_EXECUTION_ERROR_MESSAGE,
        ]
        meta_no_nones = {
            k: v
            for k, v in self.metadata.items()
            if v is not None and k not in items_to_ignore
        }
        return (self.action_name, self.action_text, jsonpickle.encode(meta_no_nones))

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return "ActionExecuted(action: {}, policy: {}, confidence: {})".format(
            self.action_name, self.policy, self.confidence
        )

    def __str__(self) -> Text:
        """Returns event as human readable string."""
        return str(self.action_name) or str(self.action_text)

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.__members__())

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, ActionExecuted):
            return NotImplemented

        return self.__members__() == other.__members__()

    def as_story_string(self) -> Optional[Text]:
        """Returns event in Markdown format."""
        if self.action_text:
            raise UnsupportedFeatureException(
                f"Printing end-to-end bot utterances is not supported in the "
                f"Markdown training format. Please use the YAML training data format "
                f"instead. Please see {DOCS_URL_TRAINING_DATA} for more information."
            )

        return self.action_name

    @classmethod
    def _from_story_string(
        cls, parameters: Dict[Text, Any]
    ) -> Optional[List["ActionExecuted"]]:
        return [
            ActionExecuted(
                parameters.get("name"),
                parameters.get("policy"),
                parameters.get("confidence"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
                parameters.get("action_text"),
                parameters.get("hide_rule_turn", False),
            )
        ]

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update(
            {
                "name": self.action_name,
                "policy": self.policy,
                "confidence": self.confidence,
                "action_text": self.action_text,
                "hide_rule_turn": self.hide_rule_turn,
            }
        )
        return d

    def as_sub_state(self) -> Dict[Text, Text]:
        """Turns ActionExecuted into a dictionary containing action name or action text.

        One action cannot have both set at the same time

        Returns:
            a dictionary containing action name or action text with the corresponding
            key.
        """
        if self.action_name:
            return {ACTION_NAME: self.action_name}
        else:
            # FIXME: we should define the type better here, and require either
            #        `action_name` or `action_text`
            return {ACTION_TEXT: cast(Text, self.action_text)}

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker.set_latest_action(self.as_sub_state())
        tracker.clear_followup_action()


class AgentUttered(SkipEventInMDStoryMixin):
    """The agent has said something to the user.

    This class is not used in the story training as it is contained in the
    ``ActionExecuted`` class. An entry is made in the ``Tracker``.
    """

    type_name = "agent"

    def __init__(
        self,
        text: Optional[Text] = None,
        data: Optional[Any] = None,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """See docstring of `BotUttered`."""
        self.text = text
        self.data = data
        super().__init__(timestamp, metadata)

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash((self.text, jsonpickle.encode(self.data)))

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, AgentUttered):
            return NotImplemented

        return (self.text, jsonpickle.encode(self.data)) == (
            other.text,
            jsonpickle.encode(other.data),
        )

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return "AgentUttered(text: {}, data: {})".format(
            self.text, json.dumps(self.data)
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update({"text": self.text, "data": self.data})
        return d

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "AgentUttered":
        try:
            return AgentUttered(
                parameters.get("text"),
                parameters.get("data"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse agent uttered event. {e}")


class ActiveLoop(Event):
    """If `name` is given: activates a loop with `name` else deactivates active loop."""

    type_name = "active_loop"

    def __init__(
        self,
        name: Optional[Text],
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event for active loop.

        Args:
            name: Name of activated loop or `None` if current loop is deactivated.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.name = name
        super().__init__(timestamp, metadata)

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return f"Loop({self.name})"

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"ActiveLoop({self.name}, {self.timestamp}, {self.metadata})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, ActiveLoop):
            return NotImplemented

        return self.name == other.name

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        props = json.dumps({LOOP_NAME: self.name})
        return f"{ActiveLoop.type_name}{props}"

    @classmethod
    def _from_story_string(cls, parameters: Dict[Text, Any]) -> List["ActiveLoop"]:
        """Called to convert a parsed story line into an event."""
        return [
            ActiveLoop(
                parameters.get(LOOP_NAME),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        ]

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update({LOOP_NAME: self.name})
        return d

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker.change_loop_to(self.name)


class LegacyForm(ActiveLoop):
    """Legacy handler of old `Form` events.

    The `ActiveLoop` event used to be called `Form`. This class is there to handle old
    legacy events which were stored with the old type name `form`.
    """

    type_name = "form"

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        # Dump old `Form` events as `ActiveLoop` events instead of keeping the old
        # event type.
        d["event"] = ActiveLoop.type_name
        return d

    def fingerprint(self) -> Text:
        """Returns the hash of the event."""
        d = self.as_dict()
        # Revert event name to legacy subclass name to avoid different event types
        # having the same fingerprint.
        d["event"] = self.type_name
        del d["timestamp"]
        return rasa.shared.utils.io.get_dictionary_fingerprint(d)


class LoopInterrupted(SkipEventInMDStoryMixin):
    """Event added by FormPolicy and RulePolicy.

    Notifies form action whether or not to validate the user input.
    """

    type_name = "loop_interrupted"

    def __init__(
        self,
        is_interrupted: bool,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Event to notify that loop was interrupted.

        This e.g. happens when a user is within a form, and is de-railing the
        form-filling by asking FAQs.

        Args:
            is_interrupted: `True` if the loop execution was interrupted, and ML
                policies had to take over the last prediction.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        super().__init__(timestamp, metadata)
        self.is_interrupted = is_interrupted

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return f"{LoopInterrupted.__name__}({self.is_interrupted})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.is_interrupted)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, LoopInterrupted):
            return NotImplemented

        return self.is_interrupted == other.is_interrupted

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "LoopInterrupted":
        return LoopInterrupted(
            parameters.get(LOOP_INTERRUPTED, False),
            parameters.get("timestamp"),
            parameters.get("metadata"),
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update({LOOP_INTERRUPTED: self.is_interrupted})
        return d

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker.interrupt_loop(self.is_interrupted)


class LegacyFormValidation(LoopInterrupted):
    """Legacy handler of old `FormValidation` events.

    The `LoopInterrupted` event used to be called `FormValidation`. This class is there
    to handle old legacy events which were stored with the old type name
    `form_validation`.
    """

    type_name = "form_validation"

    def __init__(
        self,
        validate: bool,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """See parent class docstring."""
        # `validate = True` is the same as `interrupted = False`
        super().__init__(not validate, timestamp, metadata)

    @classmethod
    def _from_parameters(cls, parameters: Dict) -> "LoopInterrupted":
        return LoopInterrupted(
            # `validate = True` means `is_interrupted = False`
            not parameters.get("validate", True),
            parameters.get("timestamp"),
            parameters.get("metadata"),
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        # Dump old `Form` events as `ActiveLoop` events instead of keeping the old
        # event type.
        d["event"] = LoopInterrupted.type_name
        return d

    def fingerprint(self) -> Text:
        """Returns hash of the event."""
        d = self.as_dict()
        # Revert event name to legacy subclass name to avoid different event types
        # having the same fingerprint.
        d["event"] = self.type_name
        del d["timestamp"]
        return rasa.shared.utils.io.get_dictionary_fingerprint(d)


class ActionExecutionRejected(SkipEventInMDStoryMixin):
    """Notify Core that the execution of the action has been rejected."""

    type_name = "action_execution_rejected"

    def __init__(
        self,
        action_name: Text,
        policy: Optional[Text] = None,
        confidence: Optional[float] = None,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event.

        Args:
            action_name: Action which was rejected.
            policy: Policy which predicted the rejected action.
            confidence: Confidence with which the reject action was predicted.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.action_name = action_name
        self.policy = policy
        self.confidence = confidence
        super().__init__(timestamp, metadata)

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return (
            "ActionExecutionRejected("
            "action: {}, policy: {}, confidence: {})"
            "".format(self.action_name, self.policy, self.confidence)
        )

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.action_name)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, ActionExecutionRejected):
            return NotImplemented

        return self.action_name == other.action_name

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "ActionExecutionRejected":
        return ActionExecutionRejected(
            parameters.get("name"),
            parameters.get("policy"),
            parameters.get("confidence"),
            parameters.get("timestamp"),
            parameters.get("metadata"),
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        d = super().as_dict()
        d.update(
            {
                "name": self.action_name,
                "policy": self.policy,
                "confidence": self.confidence,
            }
        )
        return d

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        tracker.reject_action(self.action_name)


class SessionStarted(AlwaysEqualEventMixin):
    """Mark the beginning of a new conversation session."""

    type_name = "session_started"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124320)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"SessionStarted(type_name: {self.type_name})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.type_name})"

    def as_story_string(self) -> None:
        """Skips representing event in stories."""
        logger.warning(
            f"'{self.type_name}' events cannot be serialised as story strings."
        )

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        # noinspection PyProtectedMember
        tracker._reset()


class RoutingSessionEnded(AlwaysEqualEventMixin):
    """Mark the end of a routing session for coexistence."""

    type_name = "routing_session_ended"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124434)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"RoutingSessionEnded(type_name: {self.type_name})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.type_name})"

    def as_story_string(self) -> None:
        """Skips representing event in stories."""
        logger.warning(
            f"'{self.type_name}' events cannot be serialised as story strings."
        )

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        # noinspection PyProtectedMember
        tracker._reset(is_coexistence_reset=True)


class FlowStarted(SkipEventInMDStoryMixin):
    """Mark the beginning of a new flow."""

    type_name = "flow_started"

    def __init__(
        self,
        flow_id: Text,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event to start flow.

        Args:
            flow_id: ID of the flow to be started.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.flow_id = flow_id
        super().__init__(timestamp, metadata)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"FlowStarted(flow: {self.flow_id})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.flow_id})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.flow_id)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, FlowStarted):
            return NotImplemented

        return self.flow_id == other.flow_id

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        serialized = super().as_dict()
        serialized.update({"flow_id": self.flow_id})
        return serialized

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "FlowStarted":
        try:
            return FlowStarted(
                parameters.get("flow_id"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse flow_started event. {e}")


class FlowInterrupted(SkipEventInMDStoryMixin):
    """Mark the interruption of a flow."""

    type_name = "flow_interrupted"

    def __init__(
        self,
        flow_id: Text,
        step_id: Text,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event to interrupt a flow.

        Args:
            flow_id: Name of the flow to be interrupted.
            step_id: ID of the flow step where the flow was interrupted.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.flow_id = flow_id
        self.step_id = step_id
        self.is_interrupted = True
        super().__init__(timestamp, metadata)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"FlowInterrupted(flow: {self.flow_id}, step_id: {self.step_id})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.flow_id}, {self.step_id})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(
            (
                self.flow_id,
                self.step_id,
                json.dumps({"is_interrupted": self.is_interrupted}),
            )
        )

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, FlowInterrupted):
            return NotImplemented

        return (self.flow_id, self.step_id) == (other.flow_id, other.step_id)

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        serialized = super().as_dict()
        serialized.update(
            {
                "flow_id": self.flow_id,
                "step_id": self.step_id,
            }
        )
        return serialized

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "FlowInterrupted":
        try:
            return FlowInterrupted(
                parameters.get("flow_id"),
                parameters.get("step_id"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse flow_interrupted event. {e}")


class FlowResumed(SkipEventInMDStoryMixin):
    """Mark the resuming of a flow."""

    type_name = "flow_resumed"

    def __init__(
        self,
        flow_id: Text,
        step_id: Text,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event to resume a flow.

        Args:
            flow_id: ID of the flow to be resumed.
            step_id: ID of the flow step where the flow was resumed.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.flow_id = flow_id
        self.step_id = step_id
        self.is_resumed = True
        super().__init__(timestamp, metadata)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"FlowResumed(flow: {self.flow_id}, step_id: {self.step_id})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.flow_id}, {self.step_id})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(
            (self.flow_id, self.step_id, json.dumps({"is_resumed": self.is_resumed}))
        )

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, FlowResumed):
            return NotImplemented

        return (self.flow_id, self.step_id) == (other.flow_id, other.step_id)

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        serialized = super().as_dict()
        serialized.update({"flow_id": self.flow_id, "step_id": self.step_id})
        return serialized

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "FlowResumed":
        try:
            return FlowResumed(
                parameters.get("flow_id"),
                parameters.get("step_id"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse flow_resumed event. {e}")


class FlowCompleted(SkipEventInMDStoryMixin):
    """Mark the completion of a flow."""

    type_name = "flow_completed"

    def __init__(
        self,
        flow_id: Text,
        step_id: Text,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event to complete a flow.

        Args:
            flow_id: ID of the flow to be completed.
            step_id: ID of the flow step where the flow was completed.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.flow_id = flow_id
        self.step_id = step_id
        self.is_completed = True
        super().__init__(timestamp, metadata)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"FlowCompleted(flow: {self.flow_id}, step_id: {self.step_id})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.flow_id}, {self.step_id})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(
            (
                self.flow_id,
                self.step_id,
                json.dumps({"is_completed": self.is_completed}),
            )
        )

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, FlowCompleted):
            return NotImplemented

        return (self.flow_id, self.step_id) == (other.flow_id, other.step_id)

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        serialized = super().as_dict()
        serialized.update(
            {
                "flow_id": self.flow_id,
                "step_id": self.step_id,
            }
        )
        return serialized

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "FlowCompleted":
        try:
            return FlowCompleted(
                parameters.get("flow_id"),
                parameters.get("step_id"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse flow_completed event. {e}")


class FlowCancelled(SkipEventInMDStoryMixin):
    """Mark the cancellation of a flow."""

    type_name = "flow_cancelled"

    def __init__(
        self,
        flow_id: Text,
        step_id: Text,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Creates event to cancel a flow.

        Args:
            flow_id: ID of the flow which was cancelled.
            step_id: ID of the flow step where the flow was cancelled.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.flow_id = flow_id
        self.step_id = step_id
        self.is_cancelled = True
        super().__init__(timestamp, metadata)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"FlowCancelled(flow: {self.flow_id}, step_id: {self.step_id})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.flow_id}, {self.step_id})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(
            (
                self.flow_id,
                self.step_id,
                json.dumps({"is_cancelled": self.is_cancelled}),
            )
        )

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, FlowCancelled):
            return NotImplemented

        return (self.flow_id, self.step_id) == (other.flow_id, other.step_id)

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        serialized = super().as_dict()
        serialized.update(
            {
                "flow_id": self.flow_id,
                "step_id": self.step_id,
            }
        )
        return serialized

    @classmethod
    def _from_parameters(cls, parameters: Dict[Text, Any]) -> "FlowCancelled":
        try:
            return FlowCancelled(
                parameters.get("flow_id"),
                parameters.get("step_id"),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        except KeyError as e:
            raise ValueError(f"Failed to parse flow_cancelled event. {e}")


class SessionEnded(AlwaysEqualEventMixin):
    """Mark the end of a conversation session."""

    type_name = "session_ended"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(32143124321)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"SessionEnded(type_name: {self.type_name})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.type_name})"

    def as_story_string(self) -> None:
        """Skips representing event in stories."""
        logger.warning(
            f"'{self.type_name}' events cannot be serialised as story strings."
        )

    def apply_to(self, tracker: "DialogueStateTracker") -> None:
        """Applies event to current conversation state."""
        # noinspection PyProtectedMember
        tracker._reset()


class ErrorHandled(Event):
    """An error occurred during the conversation.

    The error message is stored in the metadata of the event.
    """

    type_name = "error"

    def __init__(
        self,
        error_code: int,
        timestamp: Optional[float] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Creates event for an error.

        Args:
            error_code: Error int code.
            timestamp: When the event was created.
            metadata: Additional event metadata.
        """
        self.error_code = error_code
        super().__init__(timestamp, metadata)

    def __str__(self) -> Text:
        """Returns text representation of event."""
        return f"ErrorHandled({self.error_code})"

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"ErrorHandled({self.error_code}, {self.timestamp}, {self.metadata})"

    def __hash__(self) -> int:
        """Returns unique hash for event."""
        return hash(self.error_code)

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, ErrorHandled):
            return NotImplemented

        return self.error_code == other.error_code

    def as_story_string(self) -> Text:
        """Returns text representation of event."""
        props = json.dumps({ERROR_CODE_KEY: self.error_code})
        return f"{ErrorHandled.type_name}{props}"

    @classmethod
    def _from_story_string(cls, parameters: Dict[Text, Any]) -> List["ErrorHandled"]:
        """Called to convert a parsed story line into an event."""
        return [
            ErrorHandled(
                parameters.get(ERROR_CODE_KEY),
                parameters.get("timestamp"),
                parameters.get("metadata"),
            )
        ]

    def as_dict(self) -> Dict[Text, Any]:
        """Returns serialized event."""
        serialized = super().as_dict()
        serialized.update({ERROR_CODE_KEY: self.error_code})
        return serialized



import asyncio
import copy
import hmac
import json
import uuid
from collections import defaultdict
from dataclasses import asdict
from datetime import datetime, timedelta, timezone
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Text,
    Tuple,
    Union,
)

import structlog
from jsonschema import ValidationError, validate
from sanic import Blueprint, response
from sanic.exceptions import NotFound, SanicException, ServerError
from sanic.request import Request
from sanic.response import HTTPResponse

from rasa.core import jobs
from rasa.core.channels.channel import InputChannel, OutputChannel, UserMessage
from rasa.core.channels.voice_ready.utils import (
    CallParameters,
    validate_voice_license_scope,
)
from rasa.shared.constants import INTENT_MESSAGE_PREFIX
from rasa.shared.core.constants import USER_INTENT_SESSION_START
from rasa.shared.exceptions import RasaException
from rasa.utils.io import remove_emojis

structlogger = structlog.get_logger()

CHANNEL_NAME = "audiocodes"
KEEP_ALIVE_SECONDS = 120
KEEP_ALIVE_EXPIRATION_FACTOR = 1.5
EVENT_START = "start"
EVENT_DTMF = "DTMF"
ACTIVITY_MESSAGE = "message"
ACTIVITY_EVENT = "event"
INFO_UNKNOWN = "unknown"
ACTIVITY_ID_KEY = "id"
CREDENTIALS_TOKEN_KEY = "token"
CREDENTIALS_USE_WEBSOCKET_KEY = "use_websocket"
CREDENTIALS_KEEP_ALIVE_KEY = "keep_alive"
CREDENTIALS_KEEP_ALIVE_EXPIRATION_FACTOR_KEY = "keep_alive_expiration_factor"
CLEANUP_INTERVAL_MINUTES = 10


def map_call_params(parameters: Dict[Text, Any]) -> CallParameters:
    """Map the Audiocodes parameters to the CallParameters dataclass."""
    return CallParameters(
        call_id=parameters.get("vaigConversationId"),
        user_phone=parameters.get("caller"),
        bot_phone=parameters.get("callee"),
        user_name=parameters.get("callerDisplayName"),
        user_host=parameters.get("callerHost"),
        bot_host=parameters.get("calleeHost"),
    )


class HttpUnauthorized(SanicException):
    """**Status**: 401 Not Authorized."""

    status_code = 401
    quiet = True


class Conversation:
    def __init__(self, conversation_id: Text):
        self.activity_ids: List[Text] = []
        self.ws: Any = None
        self.conversation_id: Text = conversation_id
        self.update()

    def update(self) -> None:
        """Update the last activity time."""
        self.last_activity: datetime = datetime.now(timezone.utc)

    @staticmethod
    def get_metadata(activity: Dict[Text, Any]) -> Optional[Dict[Text, Any]]:
        """Get metadata from the activity.

        ONLY used for activities NOT for events (see _handle_event)."""
        return activity.get("parameters")

    @staticmethod
    def _handle_event(event: Dict[Text, Any]) -> Tuple[Text, Dict[Text, Any]]:
        """Handle events and return a tuple of text and metadata.

        Args:
            event: The event to handle.

        Returns:
            Tuple of text and metadata.
            text is either /session_start or /vaig_event_<event_name>
            metadata is a dictionary with the event parameters.
        """
        structlogger.debug("audiocodes.handle.event", event_payload=event)
        if "name" not in event:
            structlogger.warning(
                "audiocodes.handle.event.no_name_key", event_payload=event
            )
            return "", {}

        if event["name"] == EVENT_START:
            text = f"{INTENT_MESSAGE_PREFIX}{USER_INTENT_SESSION_START}"
            metadata = asdict(map_call_params(event.get("parameters", {})))
        elif event["name"] == EVENT_DTMF:
            text = f"{INTENT_MESSAGE_PREFIX}vaig_event_DTMF"
            metadata = {"value": event["value"]}
        else:
            # handle other events described by Audiocodes
            # https://techdocs.audiocodes.com/voice-ai-connect/#VAIG_Combined/inactivity-detection.htm?TocPath=Bot%2520integration%257CReceiving%2520notifications%257C_____3
            text = f"{INTENT_MESSAGE_PREFIX}vaig_event_{event['name']}"
            metadata = {**event.get("parameters", {})}
            if "value" in event:
                metadata["value"] = event["value"]

        return text, metadata

    def is_active_conversation(self, now: datetime, delta: timedelta) -> bool:
        """Check if the conversation is active."""
        if now - self.last_activity > delta:
            structlogger.warning(
                "audiocodes.conversation.inactive", conversation=self.conversation_id
            )
            return False
        return True

    async def handle_activities(
        self,
        message: Dict[Text, Any],
        input_channel_name: str,
        output_channel: OutputChannel,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ) -> None:
        """Handle activities sent by Audiocodes."""
        structlogger.debug("audiocodes.handle.activities")
        if input_channel_name == "":
            structlogger.warning(
                "audiocodes.handle.activities.empty_input_channel_name",
                event_info=(
                    f"Audiocodes input channel name is empty "
                    f"for conversation {self.conversation_id}"
                ),
            )

        for activity in message["activities"]:
            text = None
            if activity[ACTIVITY_ID_KEY] in self.activity_ids:
                structlogger.warning(
                    "audiocodes.handle.activities.duplicate_activity",
                    activity_id=activity[ACTIVITY_ID_KEY],
                    event_info=(
                        "Audiocodes might send duplicate activities if the bot has not "
                        "responded to the previous one or responded too late. Please "
                        "consider enabling the `use_websocket` option to use"
                        " Audiocodes Asynchronous API."
                    ),
                )
                continue
            self.activity_ids.append(activity[ACTIVITY_ID_KEY])
            if activity["type"] == ACTIVITY_MESSAGE:
                text = activity["text"]
                metadata = self.get_metadata(activity)
            elif activity["type"] == ACTIVITY_EVENT:
                text, metadata = self._handle_event(activity)
            else:
                structlogger.warning(
                    "audiocodes.handle.activities.unknown_activity_type",
                    activity_type=activity["type"],
                )
                continue

            if not text:
                continue
            user_msg = UserMessage(
                text=text,
                input_channel=input_channel_name,
                output_channel=output_channel,
                sender_id=self.conversation_id,
                metadata=metadata,
            )
            try:
                await on_new_message(user_msg)
            except Exception as e:  # skipcq: PYL-W0703
                structlogger.exception(
                    "audiocodes.handle.activities.failure",
                    sender_id=self.conversation_id,
                    error=e,
                    exc_info=True,
                )

                await output_channel.send_custom_json(
                    self.conversation_id,
                    {
                        "type": "event",
                        "name": "hangup",
                        "text": "An error occurred while handling the last message.",
                    },
                )


class AudiocodesInput(InputChannel):
    @classmethod
    def name(cls) -> Text:
        return CHANNEL_NAME

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if not credentials:
            cls.raise_missing_credentials_exception()
        schema = {
            "type": "object",
            "required": ["token"],
            "properties": {
                "keep_alive": {"type": "number"},
                "keep_alive_expiration_factor": {
                    "type": "number",
                    "minimum": 1,
                },
                "use_websocket": {"type": "boolean"},
                "token": {"type": "string"},
            },
        }

        try:
            validate(instance=credentials, schema=schema)
        except ValidationError as e:
            raise RasaException(f"Invalid credentials: {e.message}")

        return cls(
            credentials.get(CREDENTIALS_TOKEN_KEY, ""),
            credentials.get(CREDENTIALS_USE_WEBSOCKET_KEY, True),
            credentials.get(CREDENTIALS_KEEP_ALIVE_KEY, KEEP_ALIVE_SECONDS),
            credentials.get(
                CREDENTIALS_KEEP_ALIVE_EXPIRATION_FACTOR_KEY,
                KEEP_ALIVE_EXPIRATION_FACTOR,
            ),
        )

    def __init__(
        self,
        token: Text,
        use_websocket: bool,
        keep_alive: int,
        keep_alive_expiration_factor: float,
    ) -> None:
        validate_voice_license_scope()
        self.conversations: Dict[Text, Conversation] = {}
        self.token = token
        self.use_websocket = use_websocket
        self.scheduler_job = None
        self.keep_alive = keep_alive
        self.keep_alive_expiration_factor = keep_alive_expiration_factor
        self.background_tasks: Dict[Text, Set[asyncio.Task]] = defaultdict(set)

    def _create_task(self, conversation_id: Text, coro: Awaitable[Any]) -> asyncio.Task:
        """Create and track an asyncio task for a conversation."""
        task: asyncio.Task = asyncio.create_task(coro)
        self.background_tasks[conversation_id].add(task)
        task.add_done_callback(
            lambda t: self.background_tasks[conversation_id].discard(t)
        )
        return task

    async def _set_scheduler_job(self) -> None:
        if self.scheduler_job:
            self.scheduler_job.remove()
        self.scheduler_job = (await jobs.scheduler()).add_job(
            self.clean_old_conversations, "interval", minutes=CLEANUP_INTERVAL_MINUTES
        )

    def _check_token(self, token: Optional[Text]) -> None:
        if not token:
            structlogger.error("audiocodes.token_not_provided")
            raise HttpUnauthorized("Authentication token required.")

        if not hmac.compare_digest(str(token), str(self.token)):
            structlogger.error("audiocodes.invalid_token", invalid_token=token)
            raise HttpUnauthorized("Invalid authentication token.")

    def _get_conversation(
        self, token: Optional[Text], conversation_id: Text
    ) -> Conversation:
        self._check_token(token)
        conversation = self.conversations.get(conversation_id)
        if not conversation:
            raise NotFound("Conversation not found")
        conversation.update()
        return conversation

    def clean_old_conversations(self) -> None:
        structlogger.debug(
            "audiocodes.clean_old_conversations", current_number=len(self.conversations)
        )
        now = datetime.now(timezone.utc)
        delta = timedelta(seconds=self.keep_alive * self.keep_alive_expiration_factor)

        # clean up conversations
        inactive = [
            conv_id
            for conv_id, conv in self.conversations.items()
            if not conv.is_active_conversation(now, delta)
        ]

        # cancel tasks and remove conversations
        for conv_id in inactive:
            for task in self.background_tasks[conv_id]:
                task.cancel()
            self.background_tasks.pop(conv_id, None)
            self.conversations.pop(conv_id, None)

    def handle_start_conversation(self, body: Dict[Text, Any]) -> Dict[Text, Any]:
        conversation_id = body["conversation"]
        if conversation_id in self.conversations:
            raise ServerError("Conversation already exists")
        structlogger.debug(
            "audiocodes.handle_start_conversation", conversation=conversation_id
        )
        self.conversations[conversation_id] = Conversation(conversation_id)
        urls = {
            "activitiesURL": f"conversation/{conversation_id}/activities",
            "disconnectURL": f"conversation/{conversation_id}/disconnect",
            "refreshURL": f"conversation/{conversation_id}/keepalive",
            "expiresSeconds": self.keep_alive,
        }
        if self.use_websocket:
            urls.update({"websocketURL": f"conversation/{conversation_id}/websocket"})
        return urls

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        ac_webhook = Blueprint("ac_webhook", __name__)

        @ac_webhook.websocket("/conversation/<conversation_id>/websocket")  # type: ignore[misc]
        async def new_client_connection(
            request: Request, ws: Any, conversation_id: Text
        ) -> None:
            """Triggered on new websocket connection."""
            if self.use_websocket is False:
                raise ConnectionRefusedError("websocket is unavailable")
            structlogger.debug(
                "audiocodes.new_client_connection", conversation=conversation_id
            )
            conversation = self._get_conversation(request.token, conversation_id)
            if conversation:
                if conversation.ws:
                    structlogger.debug(
                        "audiocodes.new_client_connection.already_connected",
                        conversation=conversation_id,
                    )
                else:
                    conversation.ws = ws

            try:
                await ws.recv()
            except Exception:
                structlogger.warning(
                    "audiocodes.new_client_connection.closed",
                    conversation=conversation_id,
                )
                if conversation:
                    conversation.ws = None

        @ac_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            """Server health route."""
            return response.json({"status": "ok"})

        @ac_webhook.route("/webhook", methods=["GET", "POST"])
        async def receive(request: Request) -> HTTPResponse:
            """Triggered on new conversations.

            Example of payload: {"conversation": <conversation_id>, id, timestamp}.
            """
            if not self.scheduler_job:
                await self._set_scheduler_job()
            self._check_token(request.token)
            if request.method == "GET":
                return response.json({"type": "ac-bot-api", "success": True})
            return response.json(self.handle_start_conversation(request.json))

        @ac_webhook.route(
            "/conversation/<conversation_id>/activities", methods=["POST"]
        )
        async def on_activities(
            request: Request, conversation_id: Text
        ) -> HTTPResponse:
            """Process activities sent by Audiocodes.

            Activities can be:
            - Messages
            - Notifications (e.g: status of success/failure of a transfer)
            - See official documentation for more examples:
            https://techdocs.audiocodes.com/voice-ai-connect/#VAIG_Combined/sending-activities.htm
            Example of payload:
            {"conversation": <conversation_id>, "activities": List[Activity]}.
            """
            structlogger.debug("audiocodes.on_activities", conversation=conversation_id)
            conversation = self._get_conversation(request.token, conversation_id)
            if conversation is None:
                structlogger.warning(
                    "audiocodes.on_activities.no_conversation", request=request.json
                )
                return response.json({})

            if self.use_websocket:
                # send an empty response for this request
                # activities are processed in the background
                # chat response is sent via the websocket
                ac_output: Union[WebsocketOutput, AudiocodesOutput] = WebsocketOutput(
                    conversation.ws, conversation_id
                )
                self._create_task(
                    conversation_id,
                    conversation.handle_activities(
                        request.json,
                        input_channel_name=self.name(),
                        output_channel=ac_output,
                        on_new_message=on_new_message,
                    ),
                )
                return response.json({})

            # without websockets, this becomes a blocking call
            # and the response is sent back to the Audiocodes server
            # after the activities are processed
            ac_output = AudiocodesOutput()
            await conversation.handle_activities(
                request.json,
                input_channel_name=self.name(),
                output_channel=ac_output,
                on_new_message=on_new_message,
            )
            return response.json(
                {
                    "conversation": conversation_id,
                    "activities": ac_output.messages,
                }
            )

        @ac_webhook.route(
            "/conversation/<conversation_id>/disconnect", methods=["POST"]
        )
        async def disconnect(request: Request, conversation_id: Text) -> HTTPResponse:
            """Triggered when the call is disconnected.

            Example of payload:
            {"conversation": <conversation_id>, "reason": Optional[Text]}.
            """
            return await self._handle_disconnect(
                request, conversation_id, on_new_message
            )

        @ac_webhook.route("/conversation/<conversation_id>/keepalive", methods=["POST"])
        async def keepalive(request: Request, conversation_id: Text) -> HTTPResponse:
            """Triggered for keeping the connection alive.

            Invoked by VoiceAI Connect every `keep_alive`
            seconds to verify the status of the conversation
            Example of payload: # {"conversation": <conversation_id>}.
            """
            self._get_conversation(request.token, conversation_id)
            return response.json({})

        return ac_webhook

    async def _handle_disconnect(
        self,
        request: Request,
        conversation_id: Text,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ) -> HTTPResponse:
        """Triggered when the call is disconnected."""
        self._get_conversation(request.token, conversation_id)
        reason = {"reason": request.json.get("reason")}
        await on_new_message(
            UserMessage(
                text=f"{INTENT_MESSAGE_PREFIX}session_end",
                output_channel=None,
                input_channel=self.name(),
                sender_id=conversation_id,
                metadata=reason,
            )
        )
        del self.conversations[conversation_id]
        structlogger.debug(
            "audiocodes.disconnect",
            conversation=conversation_id,
            request=request.json,
        )
        return response.json({})


class AudiocodesOutput(OutputChannel):
    @classmethod
    def name(cls) -> Text:
        return CHANNEL_NAME

    def __init__(self) -> None:
        super().__init__()
        self.messages: List[Dict] = []

    async def add_message(self, message: Dict) -> None:
        """Add metadata and add message.

        Message is added to the list of
        activities to be sent to the VoiceAI Connect server.
        """
        structlogger.debug(
            "audiocodes.add.message",
            class_name=self.__class__.__name__,
            message=copy.deepcopy(message.get("text", "")),
        )
        message.update(
            {
                "timestamp": datetime.now(timezone.utc).isoformat("T")[:-3] + "Z",
                "id": str(uuid.uuid4()),
            }
        )
        await self.do_add_message(message)

    async def do_add_message(self, message: Dict) -> None:
        """Send a list of activities to the VoiceAI Connect server."""
        self.messages.append(message)

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Send a text message."""
        text = remove_emojis(text)
        await self.add_message({"type": "message", "text": text})

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        raise RasaException("Images are not supported by this channel")

    async def send_attachment(
        self, recipient_id: Text, attachment: Text, **kwargs: Any
    ) -> None:
        raise RasaException("Attachments are not supported by this channel")

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Send an activity."""
        await self.add_message(json_message)

    async def hangup(self, recipient_id: Text, **kwargs: Any) -> None:
        """Indicate that the conversation should be ended."""
        await self.add_message({"type": "event", "name": "hangup"})

    async def send_text_with_buttons(
        self,
        recipient_id: str,
        text: str,
        buttons: List[Dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Uses the concise button output format for voice channels."""
        await self.send_text_with_buttons_concise(recipient_id, text, buttons, **kwargs)


class WebsocketOutput(AudiocodesOutput):
    def __init__(self, ws: Any, conversation_id: Text) -> None:
        AudiocodesOutput.__init__(self)
        self.ws = ws
        self.conversation_id = conversation_id

    async def do_add_message(self, message: Dict) -> None:
        """Send a list of activities to the VoiceAI Connect server."""
        await self.ws.send(
            json.dumps(
                {
                    "conversation": self.conversation_id,
                    "activities": [message],
                }
            )
        )



import argparse
import os
import platform
import sys
from typing import List, Optional

import structlog
from rasa_sdk import __version__ as rasa_sdk_version

import rasa.telemetry
import rasa.utils.io
import rasa.utils.licensing
import rasa.utils.tensorflow.environment as tf_env
from rasa import version
from rasa.cli import (
    data,
    evaluate,
    export,
    interactive,
    llm_fine_tuning,
    run,
    scaffold,
    shell,
    telemetry,
    test,
    train,
    visualize,
    x,
)
from rasa.cli.arguments.default_arguments import add_logging_options
from rasa.cli.utils import (
    check_if_studio_command,
    parse_last_positional_argument_as_model_path,
    warn_if_rasa_plus_package_installed,
)
from rasa.constants import MINIMUM_COMPATIBLE_VERSION
from rasa.exceptions import DetailedRasaException
from rasa.plugin import plugin_manager
from rasa.shared.exceptions import RasaException
from rasa.utils.common import configure_logging_and_warnings
from rasa.utils.log_utils import configure_structlog

structlogger = structlog.get_logger()


def create_argument_parser() -> argparse.ArgumentParser:
    """Parse all the command line arguments for the training script."""
    parser = argparse.ArgumentParser(
        prog="rasa",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Rasa command line interface. Rasa allows you to build "
        "your own conversational assistants 🤖. The 'rasa' command "
        "allows you to easily run most common commands like "
        "creating a new bot, training or evaluating models.",
    )

    parser.add_argument(
        "--version",
        action="store_true",
        default=argparse.SUPPRESS,
        help="Print installed Rasa version",
    )

    parent_parser = argparse.ArgumentParser(add_help=False)
    add_logging_options(parent_parser)
    parent_parsers = [parent_parser]

    subparsers = parser.add_subparsers(help="Rasa commands")

    scaffold.add_subparser(subparsers, parents=parent_parsers)
    run.add_subparser(subparsers, parents=parent_parsers)
    shell.add_subparser(subparsers, parents=parent_parsers)
    train.add_subparser(subparsers, parents=parent_parsers)
    interactive.add_subparser(subparsers, parents=parent_parsers)
    telemetry.add_subparser(subparsers, parents=parent_parsers)
    test.add_subparser(subparsers, parents=parent_parsers)
    visualize.add_subparser(subparsers, parents=parent_parsers)
    data.add_subparser(subparsers, parents=parent_parsers)
    export.add_subparser(subparsers, parents=parent_parsers)
    x.add_subparser(subparsers, parents=parent_parsers)
    evaluate.add_subparser(subparsers, parents=parent_parsers)
    llm_fine_tuning.add_subparser(subparsers, parent_parsers)
    plugin_manager().hook.refine_cli(
        subparsers=subparsers, parent_parsers=parent_parsers
    )

    return parser


def print_version() -> None:
    """Prints version information of rasa tooling and python."""
    from rasa.utils.licensing import get_license_expiration_date

    print(f"Rasa Pro Version  :         {version.__version__}")
    print(f"Minimum Compatible Version: {MINIMUM_COMPATIBLE_VERSION}")
    print(f"Rasa SDK Version  :         {rasa_sdk_version}")
    print(f"Python Version    :         {platform.python_version()}")
    print(f"Operating System  :         {platform.platform()}")
    print(f"Python Path       :         {sys.executable}")
    print(f"License Expires   :         {get_license_expiration_date()}")


def main(raw_arguments: Optional[List[str]] = None) -> None:
    """Run as standalone python application.

    Args:
        raw_arguments: Arguments to parse. If not provided,
            arguments will be taken from the command line.
    """
    rasa.utils.licensing.validate_license_from_env()

    warn_if_rasa_plus_package_installed()
    parse_last_positional_argument_as_model_path()
    arg_parser = create_argument_parser()
    cmdline_arguments = arg_parser.parse_args(raw_arguments)
    log_level = getattr(cmdline_arguments, "loglevel", None)
    logging_config_file = getattr(cmdline_arguments, "logging_config_file", None)
    configure_logging_and_warnings(
        log_level, logging_config_file, warn_only_once=True, filter_repeated_logs=True
    )
    # TODO: we shouldn't configure colored logging, since we are using structlog
    # for logging - should be removed as part of logs cleanup
    rasa.utils.io.configure_colored_logging(log_level)
    configure_structlog(log_level)

    tf_env.setup_tf_environment()
    tf_env.check_deterministic_ops()

    # insert current path in syspath so custom modules are found
    sys.path.insert(1, os.getcwd())

    try:
        if hasattr(cmdline_arguments, "func"):
            is_studio_command = check_if_studio_command()

            if not is_studio_command:
                result = plugin_manager().hook.configure_commandline(
                    cmdline_arguments=cmdline_arguments
                )
                endpoints_file = result[0] if result else None

            rasa.telemetry.initialize_telemetry()
            rasa.telemetry.initialize_error_reporting()
            if not is_studio_command:
                plugin_manager().hook.init_telemetry(endpoints_file=endpoints_file)
                plugin_manager().hook.init_managers(endpoints_file=endpoints_file)

            cmdline_arguments.func(cmdline_arguments)
        elif hasattr(cmdline_arguments, "version"):
            print_version()
        else:
            # user has not provided a subcommand, let's print the help
            structlogger.error("cli.no_command", event_info="No command specified.")
            arg_parser.print_help()
            sys.exit(1)
    except DetailedRasaException as exc:
        structlogger.error(
            exc.code,
            event_info=exc.info,
            **exc.ctx,
        )
        sys.exit(1)
    except RasaException as exc:
        # these are exceptions we expect to happen (e.g. invalid training data format)
        # it doesn't make sense to print a stacktrace for these if we are not in
        # debug mode
        structlogger.debug(
            "cli.exception.details",
            event_info="Failed to run CLI command due to an exception.",
            exc_info=exc,
        )
        structlogger.error("cli.exception.rasa_exception", event_info=f"{exc}")
        sys.exit(1)
    except Exception as exc:
        structlogger.error(
            "cli.exception.general_exception",
            event_info=f"{exc.__class__.__name__}: {exc}",
            exc_info=exc,
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



from typing import Any, Dict, Optional


class EmbeddingsHealthCheckMixin:
    """Mixin class that provides methods for performing embeddings health checks during
    training and inference within components.

    This mixin offers static methods that wrap the following health check functions:
    - `perform_embeddings_health_check`
    """

    @staticmethod
    def perform_embeddings_health_check(
        custom_embeddings_config: Optional[Dict[str, Any]],
        default_embeddings_config: Dict[str, Any],
        log_source_method: str,
        log_source_component: str,
    ) -> None:
        """Wraps the `perform_embeddings_health_check` function to enable
        tracing and instrumentation.
        """
        from rasa.shared.utils.health_check.health_check import (
            perform_embeddings_health_check,
        )

        perform_embeddings_health_check(
            custom_embeddings_config,
            default_embeddings_config,
            log_source_method,
            log_source_component,
        )



from rasa.core.channels.voice_stream.tts.tts_cache import TTSCache
from rasa.core.channels.voice_stream.tts.tts_engine import (
    TTSEngine,
    TTSEngineConfig,
    TTSError,
)

__all__ = ["TTSEngine", "TTSEngineConfig", "TTSError", "TTSCache"]



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import List, Optional

import structlog

from rasa.shared.constants import (
    API_BASE_CONFIG_KEY,
    API_KEY,
    API_TYPE_CONFIG_KEY,
    API_VERSION_CONFIG_KEY,
    DEPLOYMENT_CONFIG_KEY,
    EXTRA_PARAMETERS_KEY,
    MODEL_CONFIG_KEY,
    MODEL_GROUP_ID_CONFIG_KEY,
    MODEL_GROUPS_CONFIG_KEY,
    MODELS_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
)
from rasa.shared.providers.mappings import get_client_config_class_from_provider

structlogger = structlog.get_logger()


@dataclass
class ModelConfig:
    """Parses the model config.

    Raises:
        ValueError: If the provider config key is missing in the config.
    """

    provider: str
    model: Optional[str] = None
    deployment: Optional[str] = None
    api_base: Optional[str] = None
    api_key: Optional[str] = None
    api_version: Optional[str] = None
    extra_parameters: dict = field(default_factory=dict)
    # Retained for backward compatibility with older configurations,
    # but intentionally not included in extra_parameters
    api_type: Optional[str] = None

    @classmethod
    def from_dict(cls, config: dict) -> ModelConfig:
        """Initializes a dataclass from the passed config. The provider config param is
        used to determine the client config class to use. The client config class takes
        care of resolving config aliases and throwing deprecation warnings.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Config is missing required keys.

        Returns:
            ModelConfig
        """
        from rasa.shared.utils.llm import get_provider_from_config

        # Get the provider from config, this also inferring the provider from
        # deprecated configurations
        provider = get_provider_from_config(config)

        # Retrieve the client configuration class for the specified provider.
        client_config_clazz = get_client_config_class_from_provider(provider)

        # Try to instantiate the config object in order to resolve deprecated
        # aliases and throw deprecation warnings.
        client_config_obj = client_config_clazz.from_dict(config)

        # Convert back to dictionary and instantiate the ModelConfig object.
        client_config = client_config_obj.to_dict()

        # Check for provider after resolving all aliases
        if PROVIDER_CONFIG_KEY not in client_config:
            raise ValueError(
                f"Missing required key '{PROVIDER_CONFIG_KEY}' in "
                f"'{MODELS_CONFIG_KEY}' config."
            )

        return ModelConfig(
            provider=client_config.pop(PROVIDER_CONFIG_KEY, None),
            model=client_config.pop(MODEL_CONFIG_KEY, None),
            deployment=client_config.pop(DEPLOYMENT_CONFIG_KEY, None),
            api_type=client_config.pop(API_TYPE_CONFIG_KEY, None),
            api_base=client_config.pop(API_BASE_CONFIG_KEY, None),
            api_key=client_config.pop(API_KEY, None),
            api_version=client_config.pop(API_VERSION_CONFIG_KEY, None),
            extra_parameters=client_config,
        )

    def to_dict(self) -> dict:
        """Converts the config instance into a dictionary."""
        d = asdict(self)

        # Extra parameters should also be on the top level
        d.pop(EXTRA_PARAMETERS_KEY, None)
        d.update(self.extra_parameters)

        # Remove keys with None values
        return {key: value for key, value in d.items() if value is not None}


@dataclass
class ModelGroupConfig:
    """Parses the models config. The models config is a list of model configs.

    Raises:
        ValueError: If the model group ID is None or if the models list is empty.
    """

    model_group_id: str
    models: List[ModelConfig]

    def __post_init__(self) -> None:
        if self.model_group_id is None:
            message = "Model group ID cannot be set to None."
            structlogger.error(
                "model_group_config.validation_error",
                message=message,
                model_group_id=self.model_group_id,
            )
            raise ValueError(message)
        if not self.models:
            message = "Models cannot be empty."
            structlogger.error(
                "model_group_config.validation_error",
                message=message,
                model_group_id=self.model_group_id,
            )
            raise ValueError(message)

    @classmethod
    def from_dict(cls, config: dict) -> ModelGroupConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Config is missing required keys.

        Returns:
            ModelGroupConfig
        """
        if MODELS_CONFIG_KEY not in config:
            raise ValueError(
                f"Missing required key '{MODELS_CONFIG_KEY}' in "
                f"'{MODEL_GROUPS_CONFIG_KEY}' config."
            )

        models_config = [
            ModelConfig.from_dict(model_config)
            for model_config in config[MODELS_CONFIG_KEY]
        ]

        return cls(
            model_group_id=config.get(MODEL_GROUP_ID_CONFIG_KEY),
            models=models_config,
        )

    def to_dict(self) -> dict:
        """Converts the config instance into a dictionary."""
        d = {
            MODEL_GROUP_ID_CONFIG_KEY: self.model_group_id,
            MODELS_CONFIG_KEY: [model.to_dict() for model in self.models],
        }
        return d



STUDIO_CONFIG_KEY = "studio"

KEYCLOAK_ACCESS_TOKEN_KEY = "access_token"
KEYCLOAK_EXPIRES_IN_KEY = "expires_in"
KEYCLOAK_REFRESH_EXPIRES_IN_KEY = "refresh_expires_in"
KEYCLOAK_REFRESH_TOKEN = "refresh_token"
KEYCLOAK_TOKEN_TYPE = "token_type"

RASA_STUDIO_AUTH_SERVER_URL_ENV = "RASA_STUDIO_AUTH_SERVER_URL"
RASA_STUDIO_CLI_STUDIO_URL_ENV = "RASA_STUDIO_CLI_STUDIO_URL"
RASA_STUDIO_CLI_REALM_NAME_KEY_ENV = "RASA_STUDIO_CLI_REALM_NAME_KEY"
RASA_STUDIO_CLI_CLIENT_ID_KEY_ENV = "RASA_STUDIO_CLI_CLIENT_ID_KEY"
RASA_STUDIO_CLI_DISABLE_VERIFY_KEY_ENV = "RASA_STUDIO_CLI_DISABLE_VERIFY_KEY"

STUDIO_NLU_FILENAME = "studio_nlu.yml"
STUDIO_DOMAIN_FILENAME = "studio_domain.yml"
DOMAIN_FILENAME = "domain.yml"
STUDIO_FLOWS_FILENAME = "studio_flows.yml"
STUDIO_CONFIG_FILENAME = "studio_config.yml"
STUDIO_ENDPOINTS_FILENAME = "studio_endpoints.yml"



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.db import get_contacts, write_contacts


class RemoveContact(Action):
    def name(self) -> str:
        return "remove_contact"

    def run(
        self, dispatcher: CollectingDispatcher, tracker: Tracker, domain: Dict[str, Any]
    ) -> List[Dict[Text, Any]]:
        contacts = get_contacts(tracker.sender_id)
        handle = tracker.get_slot("remove_contact_handle")

        if handle is not None:
            contact_indices_with_handle = [
                i for i, c in enumerate(contacts) if c.handle == handle
            ]
            if len(contact_indices_with_handle) == 0:
                return [SlotSet("return_value", "not_found")]
            else:
                removed_contact = contacts.pop(contact_indices_with_handle[0])
                write_contacts(tracker.sender_id, contacts)
                return [
                    SlotSet("return_value", "success"),
                    SlotSet("remove_contact_name", removed_contact.name),
                ]

        else:
            return [SlotSet("return_value", "missing_handle")]



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text, Union

import rasa.shared.constants
from rasa.core.persistor import StorageType

# WARNING: Be careful about adding any top level imports at this place!
#   These functions are imported in `rasa.__init__` and any top level import
#   added here will get executed as soon as someone runs `import rasa`.
#   Some imports are very slow (e.g. `tensorflow`) and we want them to get
#   imported when running `import rasa`. If you add more imports here,
#   please check that in the chain you are importing, no slow packages
#   are getting imported.

if TYPE_CHECKING:
    from rasa.model_training import TrainingResult
    from rasa.shared.importers.importer import TrainingDataImporter


def run(
    model: Text,
    endpoints: Text,
    connector: Optional[Text] = None,
    credentials: Optional[Text] = None,
    **kwargs: Dict[Text, Any],
) -> None:
    """Runs a Rasa model.

    Args:
        model: Path to model archive.
        endpoints: Path to endpoints file.
        connector: Connector which should be use (overwrites `credentials`
        field).
        credentials: Path to channel credentials file.
        **kwargs: Additional arguments which are passed to
        `rasa.core.run.serve_application`.

    """
    import rasa.core.run
    import rasa.shared.utils.common
    from rasa.core.available_endpoints import AvailableEndpoints
    from rasa.shared.constants import DOCS_BASE_URL
    from rasa.shared.utils.cli import print_warning

    _endpoints = AvailableEndpoints.get_instance(endpoints)

    if not connector and not credentials:
        connector = "rest"

        print_warning(
            f"No chat connector configured, falling back to the "
            f"REST input channel. To connect your bot to another channel, "
            f"read the docs here: {DOCS_BASE_URL}/messaging-and-voice-channels"
        )

    kwargs = rasa.shared.utils.common.minimal_kwargs(
        kwargs, rasa.core.run.serve_application
    )
    rasa.core.run.serve_application(
        model,
        channel=connector,
        credentials=credentials,
        endpoints=_endpoints,
        **kwargs,
    )


def train(
    domain: Text,
    config: Text,
    training_files: "Union[Text, List[Text]]",
    output: Text = rasa.shared.constants.DEFAULT_MODELS_PATH,
    dry_run: bool = False,
    force_training: bool = False,
    fixed_model_name: Optional[Text] = None,
    persist_nlu_training_data: bool = False,
    core_additional_arguments: Optional[Dict] = None,
    nlu_additional_arguments: Optional[Dict] = None,
    model_to_finetune: Optional[Text] = None,
    finetuning_epoch_fraction: float = 1.0,
    remote_storage: Optional[StorageType] = None,
    file_importer: Optional["TrainingDataImporter"] = None,
    keep_local_model_copy: bool = False,
    remote_root_only: bool = False,
) -> "TrainingResult":
    """Runs Rasa Core and NLU training in `async` loop.

    Args:
        domain: Path to the domain file.
        config: Path to the config for Core and NLU.
        training_files: Paths to the training data for Core and NLU.
        output: Output path.
        dry_run: If `True` then no training will be done, and the information about
            whether the training needs to be done will be printed.
        force_training: If `True` retrain model even if data has not changed.
        fixed_model_name: Name of model to be stored.
        persist_nlu_training_data: `True` if the NLU training data should be persisted
            with the model.
        core_additional_arguments: Additional training parameters for core training.
        nlu_additional_arguments: Additional training parameters forwarded to training
            method of each NLU component.
        model_to_finetune: Optional path to a model which should be finetuned or
            a directory in case the latest trained model should be used.
        finetuning_epoch_fraction: The fraction currently specified training epochs
            in the model configuration which should be used for finetuning.
        remote_storage: Optional name of the remote storage to
            use for storing the model.
        file_importer: Instance of `TrainingDataImporter` to use for training.
            If it is not provided, a new instance will be created.
        keep_local_model_copy: If `True` the model will be stored locally even if
            remote storage is configured.
        remote_root_only: If `True`, the model will be stored in the root of the
            remote model storage.

    Returns:
        An instance of `TrainingResult`.
    """
    from rasa.model_training import train

    return asyncio.run(
        train(
            domain=domain,
            config=config,
            training_files=training_files,
            output=output,
            dry_run=dry_run,
            force_training=force_training,
            fixed_model_name=fixed_model_name,
            persist_nlu_training_data=persist_nlu_training_data,
            core_additional_arguments=core_additional_arguments,
            nlu_additional_arguments=nlu_additional_arguments,
            model_to_finetune=model_to_finetune,
            finetuning_epoch_fraction=finetuning_epoch_fraction,
            remote_storage=remote_storage,
            file_importer=file_importer,
            keep_local_model_copy=keep_local_model_copy,
            remote_root_only=remote_root_only,
        )
    )


def test(
    model: Text,
    stories: Text,
    nlu_data: Text,
    output: Text = rasa.shared.constants.DEFAULT_RESULTS_PATH,
    additional_arguments: Optional[Dict] = None,
) -> None:
    """Test a Rasa model against a set of test data.

    Args:
        model: model to test
        stories: path to the dialogue test data
        nlu_data: path to the NLU test data
        output: path to folder where all output will be stored
        additional_arguments: additional arguments for the test call
    """
    from rasa.model_testing import test_core, test_nlu

    if additional_arguments is None:
        additional_arguments = {}

    test_core(model, stories, output, additional_arguments)  # type: ignore[unused-coroutine]
    test_nlu(model, nlu_data, output, additional_arguments)  # type: ignore[unused-coroutine]



import re
from typing import List, Optional

import structlog

from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.commands.set_slot_command import SetSlotExtractor
from rasa.dialogue_understanding.utils import add_commands_to_message_parse_data
from rasa.shared.core.domain import Domain
from rasa.shared.nlu.constants import COMMANDS, SET_SLOT_COMMAND, TEXT
from rasa.shared.nlu.training_data.message import Message

structlogger = structlog.get_logger()

SET_SLOTS_PATTERN = r"(?P<slot_name>[^(),=]+)=(?P<slot_value>[^(),]+)"
MAX_NUMBER_OF_SLOTS = 10


class CommandPayloadReader:
    @staticmethod
    def unpack_regex_message(
        message: Message,
        domain: Optional[Domain] = None,
        entity_extractor_name: Optional[
            str
        ] = SetSlotExtractor.COMMAND_PAYLOAD_READER.value,
    ) -> Message:
        """Extracts commands from user text and adds them to the message."""
        user_text = message.get(TEXT).strip()

        # prevent ReDos attacks by placing a limit on the number of slots
        if CommandPayloadReader.is_above_slot_limit(user_text):
            return message

        matches = CommandPayloadReader.find_matches(user_text)
        if not matches:
            structlogger.warning("message.parsing.failed")
            return message

        return CommandPayloadReader.extract_commands_from_pattern_matches(
            message, matches, domain, entity_extractor_name
        )

    @staticmethod
    def find_matches(user_text: str) -> List[re.Match]:
        return re.findall(SET_SLOTS_PATTERN, user_text)

    @staticmethod
    def extract_commands_from_pattern_matches(
        message: Message,
        matches: List[re.Match],
        domain: Optional[Domain] = None,
        entity_extractor_name: Optional[
            str
        ] = SetSlotExtractor.COMMAND_PAYLOAD_READER.value,
    ) -> Message:
        """Extract attributes from the matches and validate them via the domain."""
        commands = []

        for match in matches:
            slot_name = match[0].strip()
            slot_value = match[1].strip()

            domain_slot_names = [slot.name for slot in domain.slots] if domain else []

            if domain and slot_name not in domain_slot_names:
                structlogger.warning(
                    "slot.not.found",
                    slot_name=slot_name,
                )
                return message

            slot_value = (
                slot_value
                if slot_value.lower() not in {"none", "null", "undefined"}
                else None
            )

            # Create new SetSlot commands from the extracted attributes.
            command = {
                "command": SET_SLOT_COMMAND,
                "name": slot_name,
                "value": slot_value,
                "extractor": entity_extractor_name,
            }
            commands.append(command)

            add_commands_to_message_parse_data(
                message,
                CommandPayloadReader.__name__,
                [Command.command_from_json(command)],
            )

            structlogger.debug(
                "slot.set.command.added",
                slot_name=slot_name,
            )

        # set the command(s) on the Message object
        message.set(COMMANDS, commands, add_to_output=True)

        return message

    @staticmethod
    def is_above_slot_limit(user_text: str) -> bool:
        """Prevent ReDoS attacks by limiting the number of slots."""
        if user_text.count("=") > MAX_NUMBER_OF_SLOTS:
            structlogger.warning(
                "too.many.slots",
                slot_limit=10,
            )
            return True
        return False



from __future__ import annotations

import copy
from dataclasses import dataclass, field
from typing import Any, Dict, List

import structlog

from rasa.shared.constants import (
    API_TYPE_CONFIG_KEY,
    DEPLOYMENT_CONFIG_KEY,
    LITELLM_PARAMS_KEY,
    MODEL_CONFIG_KEY,
    MODEL_GROUP_ID_CONFIG_KEY,
    MODEL_LIST_KEY,
    MODEL_NAME_CONFIG_KEY,
    MODELS_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    ROUTER_CONFIG_KEY,
    USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY,
)
from rasa.shared.providers._configs.model_group_config import (
    ModelConfig,
    ModelGroupConfig,
)
from rasa.shared.providers.mappings import get_prefix_from_provider
from rasa.shared.utils.llm import DEPLOYMENT_CENTRIC_PROVIDERS

structlogger = structlog.get_logger()

_LITELLM_UNSUPPORTED_KEYS = [
    PROVIDER_CONFIG_KEY,
    DEPLOYMENT_CONFIG_KEY,
    API_TYPE_CONFIG_KEY,
    USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY,
]


@dataclass
class LiteLLMRouterClientConfig:
    """Parses configuration for a LiteLLM Router client. The configuration is expected
    to be in the following format:

    {
        "id": "model_group_id",
        "models": [
            {
                "provider": "provider_name",
                "model": "model_name",
                "api_base": "api_base",
                "api_key": "api_key",
                "api_version": "api_version",
            },
            {
                "provider": "provider_name",
                "model": "model_name",
            },
        "router": {}
    }

    This configuration is converted into the LiteLLM required format:

    {
        "id": "model_group_id",
        "model_list": [
            {
                "model_name": "model_group_id",
                "litellm_params": {
                    "model": "provider_name/model_name",
                    "api_base": "api_base",
                    "api_key": "api_key",
                    "api_version": "api_version",
                },
            },
            {
                "model_name": "model_group_id",
                "litellm_params": {
                    "model": "provider_name/model_name",
                },
            },
        ],
        "router": {},
    }

    Raises:
        ValueError: If the configuration is missing required keys.
    """

    _model_group_config: ModelGroupConfig
    router: Dict[str, Any]
    _use_chat_completions_endpoint: bool = True
    extra_parameters: dict = field(default_factory=dict)

    @property
    def model_group_id(self) -> str:
        return self._model_group_config.model_group_id

    @property
    def models(self) -> List[ModelConfig]:
        return self._model_group_config.models

    @property
    def litellm_model_list(self) -> List[Dict[str, Any]]:
        return self._convert_models_to_litellm_model_list()

    @property
    def litellm_router_settings(self) -> Dict[str, Any]:
        return self._convert_router_to_litellm_router_settings()

    @property
    def use_chat_completions_endpoint(self) -> bool:
        return self._use_chat_completions_endpoint

    def __post_init__(self) -> None:
        if not self.router:
            message = "Router cannot be empty."
            structlogger.error(
                "litellm_router_client_config.validation_error",
                message=message,
                model_group_id=self._model_group_config.model_group_id,
            )
            raise ValueError(message)

    @classmethod
    def from_dict(cls, config: dict) -> LiteLLMRouterClientConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Config is missing required keys.

        Returns:
            LiteLLMRouterClientConfig
        """
        model_group_config = ModelGroupConfig.from_dict(config)

        # Copy config to avoid mutating the original
        config_copy = copy.deepcopy(config)
        # Pop the keys used by ModelGroupConfig
        config_copy.pop(MODEL_GROUP_ID_CONFIG_KEY, None)
        config_copy.pop(MODELS_CONFIG_KEY, None)
        # Get the router settings
        router_settings = config_copy.pop(ROUTER_CONFIG_KEY, {})
        # Get the use_chat_completions_endpoint setting
        use_chat_completions_endpoint = router_settings.get(
            USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY, True
        )
        # The rest is considered as extra parameters
        extra_parameters = config_copy

        this = LiteLLMRouterClientConfig(
            _model_group_config=model_group_config,
            router=router_settings,
            _use_chat_completions_endpoint=use_chat_completions_endpoint,
            extra_parameters=extra_parameters,
        )
        return this

    def to_dict(self) -> dict:
        """Converts the config instance into a dictionary."""
        d = self._model_group_config.to_dict()
        d[ROUTER_CONFIG_KEY] = self.router
        if self.extra_parameters:
            d.update(self.extra_parameters)
        return d

    def to_litellm_dict(self) -> dict:
        return {
            **self.extra_parameters,
            MODEL_GROUP_ID_CONFIG_KEY: self.model_group_id,
            MODEL_LIST_KEY: self._convert_models_to_litellm_model_list(),
            ROUTER_CONFIG_KEY: self._convert_router_to_litellm_router_settings(),
        }

    def _convert_router_to_litellm_router_settings(self) -> Dict[str, Any]:
        _router_settings_copy = copy.deepcopy(self.router)
        _router_settings_copy.pop(USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY, None)
        return _router_settings_copy

    def _convert_models_to_litellm_model_list(self) -> List[Dict[str, Any]]:
        litellm_model_list = []

        for model_config_object in self.models:
            # Convert the model config to a dict representation
            litellm_model_config = model_config_object.to_dict()

            provider = litellm_model_config[PROVIDER_CONFIG_KEY]

            # Get the litellm prefixing for the provider
            prefix = get_prefix_from_provider(provider)

            # Determine whether to use model or deployment key based on the provider.
            litellm_model_name = (
                litellm_model_config[DEPLOYMENT_CONFIG_KEY]
                if provider in DEPLOYMENT_CENTRIC_PROVIDERS
                else litellm_model_config[MODEL_CONFIG_KEY]
            )

            # Set 'model' to a provider prefixed model name e.g. openai/gpt-4
            litellm_model_config[MODEL_CONFIG_KEY] = (
                litellm_model_name
                if f"{prefix}/" in litellm_model_name
                else f"{prefix}/{litellm_model_name}"
            )

            # Remove parameters that are None and not supported by LiteLLM.
            litellm_model_config = {
                key: value
                for key, value in litellm_model_config.items()
                if key not in _LITELLM_UNSUPPORTED_KEYS and value is not None
            }

            litellm_model_list_item = {
                MODEL_NAME_CONFIG_KEY: self.model_group_id,
                LITELLM_PARAMS_KEY: litellm_model_config,
            }

            litellm_model_list.append(litellm_model_list_item)

        return litellm_model_list
